@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2005-2006 MontaVista Software, Inc.
- * Copyright (c) 2006-2013 Red Hat, Inc.
+ * Copyright (c) 2006-2018 Red Hat, Inc.
  *
  * All rights reserved.
  *
@@ -53,7 +53,7 @@
 #include <corosync/corotypes.h>
 #include <qb/qbipc_common.h>
 #include <corosync/cfg.h>
-#include <corosync/list.h>
+#include <qb/qblist.h>
 #include <corosync/mar_gen.h>
 #include <corosync/totem/totemip.h>
 #include <corosync/totem/totem.h>
@@ -77,7 +77,7 @@ enum cfg_message_req_types {
 
 #define DEFAULT_SHUTDOWN_TIMEOUT 5
 
-static struct list_head trackers_list;
+static struct qb_list_head trackers_list;
 
 /*
  * Variables controlling a requested shutdown
@@ -91,7 +91,7 @@ static int shutdown_expected;
 
 struct cfg_info
 {
-	struct list_head list;
+	struct qb_list_head list;
 	void *conn;
 	void *tracker_conn;
 	enum {SHUTDOWN_REPLY_UNKNOWN, SHUTDOWN_REPLY_YES, SHUTDOWN_REPLY_NO} shutdown_reply;
@@ -162,6 +162,10 @@ static void message_handler_req_lib_cfg_reload_config (
 	void *conn,
 	const void *msg);
 
+static void message_handler_req_lib_cfg_reopen_log_files (
+	void *conn,
+	const void *msg);
+
 /*
  * Service Handler Definition
  */
@@ -177,15 +181,15 @@ static struct corosync_lib_handler cfg_lib_engine[] =
 	},
 	{ /* 2 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_killnode,
-		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
+		.flow_control		= CS_LIB_FLOW_CONTROL_REQUIRED
 	},
 	{ /* 3 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_tryshutdown,
-		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
+		.flow_control		= CS_LIB_FLOW_CONTROL_REQUIRED
 	},
 	{ /* 4 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_replytoshutdown,
-		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
+		.flow_control		= CS_LIB_FLOW_CONTROL_REQUIRED
 	},
 	{ /* 5 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_get_node_addrs,
@@ -197,6 +201,10 @@ static struct corosync_lib_handler cfg_lib_engine[] =
 	},
 	{ /* 7 */
 		.lib_handler_fn		= message_handler_req_lib_cfg_reload_config,
+		.flow_control		= CS_LIB_FLOW_CONTROL_REQUIRED
+	},
+	{ /* 8 */
+		.lib_handler_fn		= message_handler_req_lib_cfg_reopen_log_files,
 		.flow_control		= CS_LIB_FLOW_CONTROL_NOT_REQUIRED
 	}
 };
@@ -270,7 +278,7 @@ static char *cfg_exec_init_fn (
 {
 	api = corosync_api_v1;
 
-	list_init(&trackers_list);
+	qb_list_init(&trackers_list);
 	return (NULL);
 }
 
@@ -309,7 +317,7 @@ static int send_shutdown(void)
 static void send_test_shutdown(void *only_conn, void *exclude_conn, int status)
 {
 	struct res_lib_cfg_testshutdown res_lib_cfg_testshutdown;
-	struct list_head *iter;
+	struct qb_list_head *iter;
 
 	ENTER();
 	res_lib_cfg_testshutdown.header.size = sizeof(struct res_lib_cfg_testshutdown);
@@ -322,8 +330,8 @@ static void send_test_shutdown(void *only_conn, void *exclude_conn, int status)
 		api->ipc_dispatch_send(only_conn, &res_lib_cfg_testshutdown,
 				       sizeof(res_lib_cfg_testshutdown));
 	} else {
-		for (iter = trackers_list.next; iter != &trackers_list; iter = iter->next) {
-			struct cfg_info *ci = list_entry(iter, struct cfg_info, list);
+		qb_list_for_each(iter, &trackers_list) {
+			struct cfg_info *ci = qb_list_entry(iter, struct cfg_info, list);
 
 			if (ci->conn != exclude_conn) {
 				TRACE1("sending testshutdown to %p", ci->tracker_conn);
@@ -428,9 +436,9 @@ static void remove_ci_from_shutdown(struct cfg_info *ci)
 		api->timer_delete(shutdown_timer);
 	}
 
-	if (!list_empty(&ci->list)) {
-		list_del(&ci->list);
-		list_init(&ci->list);
+	if (!qb_list_empty(&ci->list)) {
+		qb_list_del(&ci->list);
+		qb_list_init(&ci->list);
 
 		/*
 		 * Remove our option
@@ -469,7 +477,7 @@ static int cfg_lib_init_fn (void *conn)
 	struct cfg_info *ci = (struct cfg_info *)api->ipc_private_data_get (conn);
 
 	ENTER();
-	list_init(&ci->list);
+	qb_list_init(&ci->list);
 	LEAVE();
 
         return (0);
@@ -482,23 +490,8 @@ static void message_handler_req_exec_cfg_ringreenable (
         const void *message,
         unsigned int nodeid)
 {
-	const struct req_exec_cfg_ringreenable *req_exec_cfg_ringreenable
-	  = message;
-	struct res_lib_cfg_ringreenable res_lib_cfg_ringreenable;
-
 	ENTER();
-	api->totem_ring_reenable ();
-        if (api->ipc_source_is_local(&req_exec_cfg_ringreenable->source)) {
-		res_lib_cfg_ringreenable.header.id = MESSAGE_RES_CFG_RINGREENABLE;
-		res_lib_cfg_ringreenable.header.size = sizeof (struct res_lib_cfg_ringreenable);
-		res_lib_cfg_ringreenable.header.error = CS_OK;
-		api->ipc_response_send (
-			req_exec_cfg_ringreenable->source.conn,
-			&res_lib_cfg_ringreenable,
-			sizeof (struct res_lib_cfg_ringreenable));
 
-		api->ipc_refcnt_dec(req_exec_cfg_ringreenable->source.conn);
-	}
 	LEAVE();
 }
 
@@ -521,8 +514,8 @@ static void message_handler_req_exec_cfg_killnode (
 	cs_name_t reason;
 
 	ENTER();
-	log_printf(LOGSYS_LEVEL_DEBUG, "request to kill node %d(us=%d): %s",
-		req_exec_cfg_killnode->nodeid, api->totem_nodeid_get(), reason.value);
+	log_printf(LOGSYS_LEVEL_DEBUG, "request to kill node %d(us=%d)",
+		req_exec_cfg_killnode->nodeid, api->totem_nodeid_get());
         if (req_exec_cfg_killnode->nodeid == api->totem_nodeid_get()) {
 		marshall_from_mar_name_t(&reason, &req_exec_cfg_killnode->reason);
 		log_printf(LOGSYS_LEVEL_NOTICE, "Killed by node %d: %s",
@@ -600,7 +593,11 @@ static void remove_ro_entries(icmap_map_t temp_map)
 	delete_and_notify_if_changed(temp_map, "totem.transport");
 	delete_and_notify_if_changed(temp_map, "totem.cluster_name");
 	delete_and_notify_if_changed(temp_map, "quorum.provider");
-	delete_and_notify_if_changed(temp_map, "qb.ipc_type");
+	delete_and_notify_if_changed(temp_map, "system.move_to_root_cgroup");
+	delete_and_notify_if_changed(temp_map, "system.sched_rr");
+	delete_and_notify_if_changed(temp_map, "system.priority");
+	delete_and_notify_if_changed(temp_map, "system.qb_ipc_type");
+	delete_and_notify_if_changed(temp_map, "system.state_dir");
 }
 
 /*
@@ -703,6 +700,7 @@ static void message_handler_req_exec_cfg_reload_config (
 	remove_deleted_entries(temp_map, "totem.");
 	remove_deleted_entries(temp_map, "nodelist.");
 	remove_deleted_entries(temp_map, "quorum.");
+	remove_deleted_entries(temp_map, "uidgid.config.");
 
 	/* Remove entries that cannot be changed */
 	remove_ro_entries(temp_map);
@@ -750,6 +748,8 @@ static void message_handler_req_lib_cfg_ringstatusget (
 	unsigned int iface_count;
 	char **status;
 	const char *totem_ip_string;
+	char ifname[CFG_INTERFACE_NAME_MAX_LEN];
+	unsigned int iface_ids[INTERFACE_MAX];
 	unsigned int i;
 	cs_error_t res = CS_OK;
 
@@ -760,6 +760,7 @@ static void message_handler_req_lib_cfg_ringstatusget (
 
 	api->totem_ifaces_get (
 		api->totem_nodeid_get(),
+		iface_ids,
 		interfaces,
 		INTERFACE_MAX,
 		&status,
@@ -773,11 +774,17 @@ static void message_handler_req_lib_cfg_ringstatusget (
 		totem_ip_string
 		  = (const char *)api->totem_ip_print (&interfaces[i]);
 
-		if (strlen(totem_ip_string) >= CFG_INTERFACE_NAME_MAX_LEN) {
+		if (!totem_ip_string) {
+			totem_ip_string="";
+		}
+
+		/* Allow for i/f number at the start */
+		if (strlen(totem_ip_string) >= CFG_INTERFACE_NAME_MAX_LEN-3) {
 			log_printf(LOGSYS_LEVEL_ERROR, "String representation of interface %u is too long", i);
 			res = CS_ERR_NAME_TOO_LONG;
 			goto send_response;
 		}
+		snprintf(ifname, sizeof(ifname), "%d %s", iface_ids[i], totem_ip_string);
 
 		if (strlen(status[i]) >= CFG_INTERFACE_STATUS_MAX_LEN) {
 			log_printf(LOGSYS_LEVEL_ERROR, "Status string for interface %u is too long", i);
@@ -788,7 +795,7 @@ static void message_handler_req_lib_cfg_ringstatusget (
 		strcpy ((char *)&res_lib_cfg_ringstatusget.interface_status[i],
 			status[i]);
 		strcpy ((char *)&res_lib_cfg_ringstatusget.interface_name[i],
-			totem_ip_string);
+			ifname);
 	}
 
 send_response:
@@ -805,21 +812,15 @@ static void message_handler_req_lib_cfg_ringreenable (
 	void *conn,
 	const void *msg)
 {
-	struct req_exec_cfg_ringreenable req_exec_cfg_ringreenable;
-	struct iovec iovec;
-
+	struct res_lib_cfg_ringreenable res_lib_cfg_ringreenable;
 	ENTER();
-	req_exec_cfg_ringreenable.header.size =
-		sizeof (struct req_exec_cfg_ringreenable);
-	req_exec_cfg_ringreenable.header.id = SERVICE_ID_MAKE (CFG_SERVICE,
-		MESSAGE_REQ_EXEC_CFG_RINGREENABLE);
-	api->ipc_source_set (&req_exec_cfg_ringreenable.source, conn);
-	api->ipc_refcnt_inc(conn);
 
-	iovec.iov_base = (char *)&req_exec_cfg_ringreenable;
-	iovec.iov_len = sizeof (struct req_exec_cfg_ringreenable);
-
-	assert (api->totem_mcast (&iovec, 1, TOTEM_SAFE) == 0);
+	res_lib_cfg_ringreenable.header.id = MESSAGE_RES_CFG_RINGREENABLE;
+	res_lib_cfg_ringreenable.header.size = sizeof (struct res_lib_cfg_ringreenable);
+	res_lib_cfg_ringreenable.header.error = CS_ERR_NOT_SUPPORTED;
+	api->ipc_response_send (
+		conn, &res_lib_cfg_ringreenable,
+		sizeof (struct res_lib_cfg_ringreenable));
 
 	LEAVE();
 }
@@ -863,7 +864,7 @@ static void message_handler_req_lib_cfg_tryshutdown (
 {
 	struct cfg_info *ci = (struct cfg_info *)api->ipc_private_data_get (conn);
 	const struct req_lib_cfg_tryshutdown *req_lib_cfg_tryshutdown = msg;
-	struct list_head *iter;
+	struct qb_list_head *iter;
 
 	ENTER();
 
@@ -915,8 +916,8 @@ static void message_handler_req_lib_cfg_tryshutdown (
 	 */
 	shutdown_expected = 0;
 
-	for (iter = trackers_list.next; iter != &trackers_list; iter = iter->next) {
-		struct cfg_info *testci = list_entry(iter, struct cfg_info, list);
+	qb_list_for_each(iter, &trackers_list) {
+		struct cfg_info *testci = qb_list_entry(iter, struct cfg_info, list);
 		/*
 		 * It is assumed that we will allow shutdown
 		 */
@@ -1015,11 +1016,14 @@ static void message_handler_req_lib_cfg_get_node_addrs (void *conn,
 							const void *msg)
 {
 	struct totem_ip_address node_ifs[INTERFACE_MAX];
+	unsigned int iface_ids[INTERFACE_MAX];
 	char buf[PIPE_BUF];
 	char **status;
 	unsigned int num_interfaces = 0;
+	struct sockaddr_storage *ss;
 	int ret = CS_OK;
 	int i;
+	int live_addrs = 0;
 	const struct req_lib_cfg_get_node_addrs *req_lib_cfg_get_node_addrs = msg;
 	struct res_lib_cfg_get_node_addrs *res_lib_cfg_get_node_addrs = (struct res_lib_cfg_get_node_addrs *)buf;
 	unsigned int nodeid = req_lib_cfg_get_node_addrs->nodeid;
@@ -1028,18 +1032,26 @@ static void message_handler_req_lib_cfg_get_node_addrs (void *conn,
 	if (nodeid == 0)
 		nodeid = api->totem_nodeid_get();
 
-	api->totem_ifaces_get(nodeid, node_ifs, INTERFACE_MAX, &status, &num_interfaces);
+	if (api->totem_ifaces_get(nodeid, iface_ids, node_ifs, INTERFACE_MAX, &status, &num_interfaces)) {
+		ret = CS_ERR_EXIST;
+		num_interfaces = 0;
+	}
 
 	res_lib_cfg_get_node_addrs->header.size = sizeof(struct res_lib_cfg_get_node_addrs) + (num_interfaces * TOTEMIP_ADDRLEN);
 	res_lib_cfg_get_node_addrs->header.id = MESSAGE_RES_CFG_GET_NODE_ADDRS;
 	res_lib_cfg_get_node_addrs->header.error = ret;
-	res_lib_cfg_get_node_addrs->num_addrs = num_interfaces;
 	if (num_interfaces) {
 		res_lib_cfg_get_node_addrs->family = node_ifs[0].family;
 		for (i = 0, addr_buf = (char *)res_lib_cfg_get_node_addrs->addrs;
-		    i < num_interfaces; i++, addr_buf += TOTEMIP_ADDRLEN) {
-			memcpy(addr_buf, node_ifs[i].addr, TOTEMIP_ADDRLEN);
+		    i < num_interfaces; i++) {
+			ss = (struct sockaddr_storage *)&node_ifs[i].addr;
+			if (ss->ss_family) {
+				memcpy(addr_buf, node_ifs[i].addr, TOTEMIP_ADDRLEN);
+				live_addrs++;
+				addr_buf += TOTEMIP_ADDRLEN;
+			}
 		}
+		res_lib_cfg_get_node_addrs->num_addrs = live_addrs;
 	} else {
 		res_lib_cfg_get_node_addrs->header.error = CS_ERR_NOT_EXIST;
 	}
@@ -1080,3 +1092,24 @@ static void message_handler_req_lib_cfg_reload_config (void *conn, const void *m
 
 	LEAVE();
 }
+
+static void message_handler_req_lib_cfg_reopen_log_files (void *conn, const void *msg)
+{
+	struct res_lib_cfg_reopen_log_files res_lib_cfg_reopen_log_files;
+	cs_error_t res;
+
+	ENTER();
+
+	log_printf(LOGSYS_LEVEL_DEBUG, "Reopening logging files\n");
+
+	res = logsys_reopen_log_files();
+
+	res_lib_cfg_reopen_log_files.header.size = sizeof(res_lib_cfg_reopen_log_files);
+	res_lib_cfg_reopen_log_files.header.id = MESSAGE_RES_CFG_REOPEN_LOG_FILES;
+	res_lib_cfg_reopen_log_files.header.error = res;
+	api->ipc_response_send(conn,
+			       &res_lib_cfg_reopen_log_files,
+			       sizeof(res_lib_cfg_reopen_log_files));
+
+	LEAVE();
+}
