@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2002-2006 MontaVista Software, Inc.
- * Copyright (c) 2006-2012 Red Hat, Inc.
+ * Copyright (c) 2006-2018 Red Hat, Inc.
  *
  * All rights reserved.
  *
@@ -96,6 +96,10 @@
 #include <semaphore.h>
 #include <string.h>
 
+#ifdef HAVE_LIBSYSTEMD
+#include <systemd/sd-daemon.h>
+#endif
+
 #include <qb/qbdefs.h>
 #include <qb/qblog.h>
 #include <qb/qbloop.h>
@@ -105,15 +109,10 @@
 #include <corosync/swab.h>
 #include <corosync/corotypes.h>
 #include <corosync/corodefs.h>
-#include <corosync/list.h>
 #include <corosync/totem/totempg.h>
 #include <corosync/logsys.h>
 #include <corosync/icmap.h>
 
-#ifdef HAVE_LIBCGROUP
-#include <libcgroup.h>
-#endif
-
 #include "quorum.h"
 #include "totemsrp.h"
 #include "logconfig.h"
@@ -125,6 +124,8 @@
 #include "apidef.h"
 #include "service.h"
 #include "schedwrk.h"
+#include "ipcs_stats.h"
+#include "stats.h"
 
 #ifdef HAVE_SMALL_MEMORY_FOOTPRINT
 #define IPC_LOGSYS_SIZE			1024*64
@@ -164,7 +165,7 @@ static corosync_timer_handle_t corosync_stats_timer_handle;
 
 static const char *corosync_lock_file = LOCALSTATEDIR"/run/corosync.pid";
 
-static int ip_version = AF_INET;
+static char corosync_config_file[PATH_MAX + 1] = COROSYSCONFDIR "/corosync.conf";
 
 qb_loop_t *cs_poll_handle_get (void)
 {
@@ -200,6 +201,12 @@ void corosync_state_dump (void)
 	}
 }
 
+const char *corosync_get_config_file(void)
+{
+
+	return (corosync_config_file);
+}
+
 static void corosync_blackbox_write_to_file (void)
 {
 	char fname[PATH_MAX];
@@ -213,16 +220,19 @@ static void corosync_blackbox_write_to_file (void)
 	localtime_r(&cur_time_t, &cur_time_tm);
 
 	strftime(time_str, PATH_MAX, "%Y-%m-%dT%H:%M:%S", &cur_time_tm);
-	snprintf(fname, PATH_MAX, "%s/fdata-%s-%lld",
-	    get_run_dir(),
+	if (snprintf(fname, PATH_MAX, "%s/fdata-%s-%lld",
+	    get_state_dir(),
 	    time_str,
-	    (long long int)getpid());
+	    (long long int)getpid()) >= PATH_MAX) {
+		log_printf(LOGSYS_LEVEL_ERROR, "Can't snprintf blackbox file name");
+		return ;
+	}
 
 	if ((res = qb_log_blackbox_write_to_file(fname)) < 0) {
 		LOGSYS_PERROR(-res, LOGSYS_LEVEL_ERROR, "Can't store blackbox file");
 		return ;
 	}
-	snprintf(fdata_fname, sizeof(fdata_fname), "%s/fdata", get_run_dir());
+	snprintf(fdata_fname, sizeof(fdata_fname), "%s/fdata", get_state_dir());
 	unlink(fdata_fname);
 	if (symlink(fname, fdata_fname) == -1) {
 		log_printf(LOGSYS_LEVEL_ERROR, "Can't create symlink to '%s' for corosync blackbox file '%s'",
@@ -292,6 +302,10 @@ static void corosync_sync_completed (void)
 	 * Inform totem to start using new message queue again
 	 */
 	totempg_trans_ack();
+
+#ifdef HAVE_LIBSYSTEMD
+	sd_notify (0, "READY=1");
+#endif
 }
 
 static int corosync_sync_callbacks_retrieve (
@@ -324,11 +338,11 @@ static void member_object_joined (unsigned int nodeid)
 	char member_status[ICMAP_KEYNAME_MAXLEN];
 
 	snprintf(member_ip, ICMAP_KEYNAME_MAXLEN,
-		"runtime.totem.pg.mrp.srp.members.%u.ip", nodeid);
+		"runtime.members.%u.ip", nodeid);
 	snprintf(member_join_count, ICMAP_KEYNAME_MAXLEN,
-		"runtime.totem.pg.mrp.srp.members.%u.join_count", nodeid);
+		"runtime.members.%u.join_count", nodeid);
 	snprintf(member_status, ICMAP_KEYNAME_MAXLEN,
-		"runtime.totem.pg.mrp.srp.members.%u.status", nodeid);
+		"runtime.members.%u.status", nodeid);
 
 	if (icmap_get(member_ip, NULL, NULL, NULL) == CS_OK) {
 		icmap_inc(member_join_count);
@@ -348,7 +362,7 @@ static void member_object_left (unsigned int nodeid)
 	char member_status[ICMAP_KEYNAME_MAXLEN];
 
 	snprintf(member_status, ICMAP_KEYNAME_MAXLEN,
-		"runtime.totem.pg.mrp.srp.members.%u.status", nodeid);
+		"runtime.members.%u.status", nodeid);
 	icmap_set_string(member_status, "left");
 
 	log_printf (LOGSYS_LEVEL_DEBUG,
@@ -475,54 +489,24 @@ static void corosync_totem_stats_updater (void *data)
 	uint32_t total_mtt_rx_token;
 	uint32_t total_backlog_calc;
 	uint32_t total_token_holdtime;
-	int t, prev, i;
+	int t, prev;
 	int32_t token_count;
-	char key_name[ICMAP_KEYNAME_MAXLEN];
 	const char *cstr;
 
 	stats = api->totem_get_stats();
 
-	icmap_set_uint32("runtime.totem.pg.msg_reserved", stats->msg_reserved);
-	icmap_set_uint32("runtime.totem.pg.msg_queue_avail", stats->msg_queue_avail);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.orf_token_tx", stats->mrp->srp->orf_token_tx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.orf_token_rx", stats->mrp->srp->orf_token_rx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.memb_merge_detect_tx", stats->mrp->srp->memb_merge_detect_tx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.memb_merge_detect_rx", stats->mrp->srp->memb_merge_detect_rx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.memb_join_tx", stats->mrp->srp->memb_join_tx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.memb_join_rx", stats->mrp->srp->memb_join_rx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.mcast_tx", stats->mrp->srp->mcast_tx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.mcast_retx", stats->mrp->srp->mcast_retx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.mcast_rx", stats->mrp->srp->mcast_rx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.memb_commit_token_tx", stats->mrp->srp->memb_commit_token_tx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.memb_commit_token_rx", stats->mrp->srp->memb_commit_token_rx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.token_hold_cancel_tx", stats->mrp->srp->token_hold_cancel_tx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.token_hold_cancel_rx", stats->mrp->srp->token_hold_cancel_rx);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.operational_entered", stats->mrp->srp->operational_entered);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.operational_token_lost", stats->mrp->srp->operational_token_lost);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.gather_entered", stats->mrp->srp->gather_entered);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.gather_token_lost", stats->mrp->srp->gather_token_lost);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.commit_entered", stats->mrp->srp->commit_entered);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.commit_token_lost", stats->mrp->srp->commit_token_lost);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.recovery_entered", stats->mrp->srp->recovery_entered);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.recovery_token_lost", stats->mrp->srp->recovery_token_lost);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.consensus_timeouts", stats->mrp->srp->consensus_timeouts);
-	icmap_set_uint64("runtime.totem.pg.mrp.srp.rx_msg_dropped", stats->mrp->srp->rx_msg_dropped);
-	icmap_set_uint32("runtime.totem.pg.mrp.srp.continuous_gather", stats->mrp->srp->continuous_gather);
-	icmap_set_uint32("runtime.totem.pg.mrp.srp.continuous_sendmsg_failures",
-	    stats->mrp->srp->continuous_sendmsg_failures);
 
-	icmap_set_uint8("runtime.totem.pg.mrp.srp.firewall_enabled_or_nic_failure",
-		stats->mrp->srp->continuous_gather > MAX_NO_CONT_GATHER ? 1 : 0);
+	stats->srp->firewall_enabled_or_nic_failure = stats->srp->continuous_gather > MAX_NO_CONT_GATHER ? 1 : 0;
 
-	if (stats->mrp->srp->continuous_gather > MAX_NO_CONT_GATHER ||
-	    stats->mrp->srp->continuous_sendmsg_failures > MAX_NO_CONT_SENDMSG_FAILURES) {
+	if (stats->srp->continuous_gather > MAX_NO_CONT_GATHER ||
+	    stats->srp->continuous_sendmsg_failures > MAX_NO_CONT_SENDMSG_FAILURES) {
 		cstr = "";
 
-		if (stats->mrp->srp->continuous_sendmsg_failures > MAX_NO_CONT_SENDMSG_FAILURES) {
+		if (stats->srp->continuous_sendmsg_failures > MAX_NO_CONT_SENDMSG_FAILURES) {
 			cstr = "number of multicast sendmsg failures is above threshold";
 		}
 
-		if (stats->mrp->srp->continuous_gather > MAX_NO_CONT_GATHER) {
+		if (stats->srp->continuous_gather > MAX_NO_CONT_GATHER) {
 			cstr = "totem is continuously in gather state";
 		}
 
@@ -531,44 +515,43 @@ static void corosync_totem_stats_updater (void *data)
 			"operating system or network fault (reason: %s). The most common "
 			"cause of this message is that the local firewall is "
 			"configured improperly.", cstr);
-		icmap_set_uint8("runtime.totem.pg.mrp.srp.firewall_enabled_or_nic_failure", 1);
+		stats->srp->firewall_enabled_or_nic_failure = 1;
 	} else {
-		icmap_set_uint8("runtime.totem.pg.mrp.srp.firewall_enabled_or_nic_failure", 0);
+		stats->srp->firewall_enabled_or_nic_failure = 0;
 	}
 
-	for (i = 0; i < stats->mrp->srp->rrp->interface_count; i++) {
-		snprintf(key_name, ICMAP_KEYNAME_MAXLEN, "runtime.totem.pg.mrp.rrp.%u.faulty", i);
-		icmap_set_uint8(key_name, stats->mrp->srp->rrp->faulty[i]);
-	}
 	total_mtt_rx_token = 0;
 	total_token_holdtime = 0;
 	total_backlog_calc = 0;
 	token_count = 0;
-	t = stats->mrp->srp->latest_token;
+	t = stats->srp->latest_token;
 	while (1) {
 		if (t == 0)
 			prev = TOTEM_TOKEN_STATS_MAX - 1;
 		else
 			prev = t - 1;
-		if (prev == stats->mrp->srp->earliest_token)
+		if (prev == stats->srp->earliest_token)
 			break;
 		/* if tx == 0, then dropped token (not ours) */
-		if (stats->mrp->srp->token[t].tx != 0 ||
-			(stats->mrp->srp->token[t].rx - stats->mrp->srp->token[prev].rx) > 0 ) {
-			total_mtt_rx_token += (stats->mrp->srp->token[t].rx - stats->mrp->srp->token[prev].rx);
-			total_token_holdtime += (stats->mrp->srp->token[t].tx - stats->mrp->srp->token[t].rx);
-			total_backlog_calc += stats->mrp->srp->token[t].backlog_calc;
+		if (stats->srp->token[t].tx != 0 ||
+			(stats->srp->token[t].rx - stats->srp->token[prev].rx) > 0 ) {
+			total_mtt_rx_token += (stats->srp->token[t].rx - stats->srp->token[prev].rx);
+			total_token_holdtime += (stats->srp->token[t].tx - stats->srp->token[t].rx);
+			total_backlog_calc += stats->srp->token[t].backlog_calc;
 			token_count++;
 		}
 		t = prev;
 	}
 	if (token_count) {
-		icmap_set_uint32("runtime.totem.pg.mrp.srp.mtt_rx_token", (total_mtt_rx_token / token_count));
-		icmap_set_uint32("runtime.totem.pg.mrp.srp.avg_token_workload", (total_token_holdtime / token_count));
-		icmap_set_uint32("runtime.totem.pg.mrp.srp.avg_backlog_calc", (total_backlog_calc / token_count));
+		stats->srp->mtt_rx_token = (total_mtt_rx_token / token_count);
+		stats->srp->avg_token_workload = (total_token_holdtime / token_count);
+		stats->srp->avg_backlog_calc = (total_backlog_calc / token_count);
 	}
 
-	cs_ipcs_stats_update();
+	stats->srp->time_since_token_last_received = qb_util_nano_current_get () / QB_TIME_NS_IN_MSEC -
+		stats->srp->token[stats->srp->latest_token].rx;
+
+	stats_trigger_trackers();
 
 	api->timer_add_duration (1500 * MILLI_2_NANO_SECONDS, NULL,
 		corosync_totem_stats_updater,
@@ -577,10 +560,6 @@ static void corosync_totem_stats_updater (void *data)
 
 static void corosync_totem_stats_init (void)
 {
-	icmap_set_uint32("runtime.totem.pg.mrp.srp.mtt_rx_token", 0);
-	icmap_set_uint32("runtime.totem.pg.mrp.srp.avg_token_workload", 0);
-	icmap_set_uint32("runtime.totem.pg.mrp.srp.avg_backlog_calc", 0);
-
 	/* start stats timer */
 	api->timer_add_duration (1500 * MILLI_2_NANO_SECONDS, NULL,
 		corosync_totem_stats_updater,
@@ -653,14 +632,14 @@ int main_mcast (
 
 static void corosync_ring_id_create_or_load (
 	struct memb_ring_id *memb_ring_id,
-	const struct totem_ip_address *addr)
+	unsigned int nodeid)
 {
 	int fd;
 	int res = 0;
 	char filename[PATH_MAX];
 
-	snprintf (filename, sizeof(filename), "%s/ringid_%s",
-		get_run_dir(), totemip_print (addr));
+	snprintf (filename, sizeof(filename), "%s/ringid_%u",
+		get_state_dir(), nodeid);
 	fd = open (filename, O_RDONLY, 0700);
 	/*
 	 * If file can be opened and read, read the ring id
@@ -693,20 +672,19 @@ static void corosync_ring_id_create_or_load (
 		}
 	}
 
-	totemip_copy(&memb_ring_id->rep, addr);
-	assert (!totemip_zero_check(&memb_ring_id->rep));
+	memb_ring_id->rep = nodeid;
 }
 
 static void corosync_ring_id_store (
 	const struct memb_ring_id *memb_ring_id,
-	const struct totem_ip_address *addr)
+	unsigned int nodeid)
 {
 	char filename[PATH_MAX];
 	int fd;
 	int res;
 
-	snprintf (filename, sizeof(filename), "%s/ringid_%s",
-		get_run_dir(), totemip_print (addr));
+	snprintf (filename, sizeof(filename), "%s/ringid_%u",
+		get_state_dir(), nodeid);
 
 	fd = open (filename, O_WRONLY, 0700);
 	if (fd == -1) {
@@ -769,6 +747,11 @@ int corosync_sending_allowed (
 		return -EINVAL;
 	}
 
+	/* Message ID out of range */
+	if (id >= corosync_service[service]->lib_engine_count) {
+		return -EINVAL;
+	}
+
 	sending_allowed = QB_FALSE;
 	if (corosync_quorum_is_quorate() == 1 ||
 	    corosync_service[service]->allow_inquorate == CS_LIB_ALLOW_INQUORATE) {
@@ -834,7 +817,6 @@ static void timer_function_scheduler_timeout (void *data)
 	struct scheduler_pause_timeout_data *timeout_data = (struct scheduler_pause_timeout_data *)data;
 	unsigned long long tv_current;
 	unsigned long long tv_diff;
-	uint64_t schedmiss_event_tstamp;
 
 	tv_current = qb_util_nano_current_get ();
 
@@ -850,15 +832,9 @@ static void timer_function_scheduler_timeout (void *data)
 	timeout_data->tv_prev = tv_current;
 
 	if (tv_diff > timeout_data->max_tv_diff) {
-		schedmiss_event_tstamp = qb_util_nano_from_epoch_get() / QB_TIME_NS_IN_MSEC;
-
-		log_printf (LOGSYS_LEVEL_WARNING, "Corosync main process was not scheduled (@%" PRIu64 ") for %0.4f ms "
+		log_printf (LOGSYS_LEVEL_WARNING, "Corosync main process was not scheduled for %0.4f ms "
 		    "(threshold is %0.4f ms). Consider token timeout increase.",
-		    schedmiss_event_tstamp,
 		    (float)tv_diff / QB_TIME_NS_IN_MSEC, (float)timeout_data->max_tv_diff / QB_TIME_NS_IN_MSEC);
-
-		icmap_set_float("runtime.schedmiss.delay", (float)tv_diff / QB_TIME_NS_IN_MSEC);
-		icmap_set_uint64("runtime.schedmiss.timestamp", schedmiss_event_tstamp);
 	}
 
 	/*
@@ -999,6 +975,41 @@ static void corosync_fplay_control_init (void)
 			NULL, &track);
 }
 
+static void force_gather_notify_fn(
+	int32_t event,
+	const char *key_name,
+	struct icmap_notify_value new_val,
+	struct icmap_notify_value old_val,
+	void *user_data)
+{
+	char *key_val;
+
+	if (icmap_get_string(key_name, &key_val) == CS_OK && strcmp(key_val, "no") == 0)
+		goto out;
+
+	icmap_set_string("runtime.force_gather", "no");
+
+	if (strcmp(key_name, "runtime.force_gather") == 0) {
+		log_printf(LOGSYS_LEVEL_ERROR, "Forcing into GATHER state\n");
+		totempg_force_gather();
+	}
+
+out:
+	free(key_val);
+}
+
+static void corosync_force_gather_init (void)
+{
+	icmap_track_t track = NULL;
+
+	icmap_set_string("runtime.force_gather", "no");
+
+	icmap_track_add("runtime.force_gather",
+			ICMAP_TRACK_ADD | ICMAP_TRACK_DELETE | ICMAP_TRACK_MODIFY,
+			force_gather_notify_fn,
+			NULL, &track);
+}
+
 /*
  * Set RO flag for keys, which ether doesn't make sense to change by user (statistic)
  * or which when changed are not reflected by runtime (totem.crypto_cipher, ...).
@@ -1012,11 +1023,12 @@ static void set_icmap_ro_keys_flag (void)
 	 * Set RO flag for all keys of internal configuration and runtime statistics
 	 */
 	icmap_set_ro_access("internal_configuration.", CS_TRUE, CS_TRUE);
-	icmap_set_ro_access("runtime.connections.", CS_TRUE, CS_TRUE);
-	icmap_set_ro_access("runtime.totem.", CS_TRUE, CS_TRUE);
 	icmap_set_ro_access("runtime.services.", CS_TRUE, CS_TRUE);
 	icmap_set_ro_access("runtime.config.", CS_TRUE, CS_TRUE);
+	icmap_set_ro_access("runtime.totem.", CS_TRUE, CS_TRUE);
 	icmap_set_ro_access("uidgid.config.", CS_TRUE, CS_TRUE);
+	icmap_set_ro_access("system.", CS_TRUE, CS_TRUE);
+	icmap_set_ro_access("nodelist.", CS_TRUE, CS_TRUE);
 
 	/*
 	 * Set RO flag for constrete keys of configuration which can't be changed
@@ -1034,7 +1046,6 @@ static void set_icmap_ro_keys_flag (void)
 	icmap_set_ro_access("totem.version", CS_FALSE, CS_TRUE);
 	icmap_set_ro_access("totem.nodeid", CS_FALSE, CS_TRUE);
 	icmap_set_ro_access("totem.clear_node_high_bit", CS_FALSE, CS_TRUE);
-	icmap_set_ro_access("qb.ipc_type", CS_FALSE, CS_TRUE);
 	icmap_set_ro_access("config.reload_in_progress", CS_FALSE, CS_TRUE);
 	icmap_set_ro_access("config.totemconfig_reload_in_progress", CS_FALSE, CS_TRUE);
 }
@@ -1054,6 +1065,8 @@ static void main_service_ready (void)
 	cs_ipcs_init();
 	corosync_totem_stats_init ();
 	corosync_fplay_control_init ();
+	corosync_force_gather_init ();
+
 	sync_init (
 		corosync_sync_callbacks_retrieve,
 		corosync_sync_completed);
@@ -1147,84 +1160,47 @@ error_close:
 }
 
 static int corosync_move_to_root_cgroup(void) {
+	FILE *f;
 	int res = -1;
-#ifdef HAVE_LIBCGROUP
-	int cg_ret;
-	struct cgroup *root_cgroup = NULL;
-	struct cgroup_controller *root_cpu_cgroup_controller = NULL;
-	char *current_cgroup_path = NULL;
 
-	cg_ret = cgroup_init();
-	if (cg_ret) {
-		log_printf(LOGSYS_LEVEL_WARNING, "Unable to initialize libcgroup: %s ",
-		    cgroup_strerror(cg_ret));
-
-		goto exit_res;
-	}
-
-	cg_ret = cgroup_get_current_controller_path(getpid(), "cpu", &current_cgroup_path);
-	if (cg_ret) {
-		log_printf(LOGSYS_LEVEL_WARNING, "Unable to get current cpu cgroup path: %s ",
-		    cgroup_strerror(cg_ret));
-
-		goto exit_res;
-	}
-
-	if (strcmp(current_cgroup_path, "/") == 0) {
-		log_printf(LOGSYS_LEVEL_DEBUG, "Corosync is already in root cgroup path");
+	/*
+	 * /sys/fs/cgroup is hardcoded, because most of Linux distributions are now
+	 * using systemd and systemd uses hardcoded path of cgroup mount point.
+	 *
+	 * This feature is expected to be removed as soon as systemd gets support
+	 * for managing RT configuration.
+	 */
+	f = fopen("/sys/fs/cgroup/cpu/cpu.rt_runtime_us", "rt");
+	if (f == NULL) {
+		log_printf(LOGSYS_LEVEL_DEBUG, "cpu.rt_runtime_us doesn't exists -> "
+		    "system without cgroup or with disabled CONFIG_RT_GROUP_SCHED");
 
 		res = 0;
 		goto exit_res;
 	}
+	(void)fclose(f);
 
-	root_cgroup = cgroup_new_cgroup("/");
-	if (root_cgroup == NULL) {
-		log_printf(LOGSYS_LEVEL_WARNING, "Can't create root cgroup");
+	f = fopen("/sys/fs/cgroup/cpu/tasks", "w");
+	if (f == NULL) {
+		log_printf(LOGSYS_LEVEL_WARNING, "Can't open cgroups tasks file for writing");
 
 		goto exit_res;
 	}
 
-	root_cpu_cgroup_controller = cgroup_add_controller(root_cgroup, "cpu");
-	if (root_cpu_cgroup_controller == NULL) {
-		log_printf(LOGSYS_LEVEL_WARNING, "Can't create root cgroup cpu controller");
+	if (fprintf(f, "%jd\n", (intmax_t)getpid()) <= 0) {
+		log_printf(LOGSYS_LEVEL_WARNING, "Can't write corosync pid into cgroups tasks file");
 
-		goto exit_res;
+		goto close_and_exit_res;
 	}
 
-	cg_ret = cgroup_attach_task(root_cgroup);
-	if (cg_ret) {
-		log_printf(LOGSYS_LEVEL_WARNING, "Can't attach task to root cgroup: %s ",
-		    cgroup_strerror(cg_ret));
+close_and_exit_res:
+	if (fclose(f) != 0) {
+		log_printf(LOGSYS_LEVEL_WARNING, "Can't close cgroups tasks file");
 
 		goto exit_res;
 	}
 
-	cg_ret = cgroup_get_current_controller_path(getpid(), "cpu", &current_cgroup_path);
-	if (cg_ret) {
-		log_printf(LOGSYS_LEVEL_WARNING, "Unable to get current cpu cgroup path: %s ",
-		    cgroup_strerror(cg_ret));
-
-		goto exit_res;
-	}
-
-	if (strcmp(current_cgroup_path, "/") == 0) {
-		log_printf(LOGSYS_LEVEL_NOTICE, "Corosync successfully moved to root cgroup");
-		res = 0;
-	} else {
-		log_printf(LOGSYS_LEVEL_WARNING, "Can't move Corosync to root cgroup");
-	}
-
 exit_res:
-	if (root_cgroup != NULL) {
-		cgroup_free(&root_cgroup);
-	}
-
-	/*
-	 * libcgroup doesn't define something like cgroup_fini so there is no way how to clean
-	 * it's cache. It has to be called when libcgroup authors decide to implement it.
-	 */
-
-#endif
 	 return (res);
 }
 
@@ -1241,54 +1217,36 @@ int main (int argc, char **argv, char **envp)
 	struct scheduler_pause_timeout_data scheduler_pause_timeout_data;
 	long int tmpli;
 	char *ep;
+	char *tmp_str;
+	int log_subsys_id_totem;
 
 	/* default configuration
 	 */
 	background = 1;
-	sched_rr = 1;
-	prio = 0;
 	testonly = 0;
-	move_to_root_cgroup = 1;
 
-	while ((ch = getopt (argc, argv, "fP:pRrtv")) != EOF) {
+	while ((ch = getopt (argc, argv, "c:ftv")) != EOF) {
 
 		switch (ch) {
-			case 'f':
-				background = 0;
-				break;
-			case 'p':
-				sched_rr = 0;
-				break;
-			case 'P':
-				if (strcmp(optarg, "max") == 0) {
-					prio = INT_MIN;
-				} else if (strcmp(optarg, "min") == 0) {
-					prio = INT_MAX;
-				} else {
-					errno = 0;
+			case 'c':
+				res = snprintf(corosync_config_file, sizeof(corosync_config_file), "%s", optarg);
+				if (res >= sizeof(corosync_config_file)) {
+					fprintf (stderr, "Config file path too long.\n");
+					syslog (LOGSYS_LEVEL_ERROR, "Config file path too long.");
 
-					tmpli = strtol(optarg, &ep, 10);
-					if (errno != 0 || *ep != '\0' || tmpli > INT_MAX || tmpli < INT_MIN) {
-						fprintf(stderr, "Priority value %s is invalid", optarg);
-						logsys_system_fini();
-						return EXIT_FAILURE;
-					}
-
-					prio = tmpli;
+					logsys_system_fini();
+					return EXIT_FAILURE;
 				}
 				break;
-			case 'R':
-				move_to_root_cgroup = 0;
-				break;
-			case 'r':
-				sched_rr = 1;
+			case 'f':
+				background = 0;
 				break;
 			case 't':
 				testonly = 1;
 				break;
 			case 'v':
 				printf ("Corosync Cluster Engine, version '%s'\n", VERSION);
-				printf ("Copyright (c) 2006-2009 Red Hat, Inc.\n");
+				printf ("Copyright (c) 2006-2018 Red Hat, Inc.\n");
 				logsys_system_fini();
 				return EXIT_SUCCESS;
 
@@ -1296,11 +1254,8 @@ int main (int argc, char **argv, char **envp)
 			default:
 				fprintf(stderr, \
 					"usage:\n"\
+					"        -c     : Corosync config file path.\n"\
 					"        -f     : Start application in foreground.\n"\
-					"        -p     : Do not set realtime scheduling.\n"\
-					"        -r     : Set round robin realtime scheduling (default).\n"\
-					"        -R     : Do not try move corosync to root cpu cgroup (valid when built with libcgroup)\n" \
-					"        -P num : Set priority of process (no effect when -r is used)\n"\
 					"        -t     : Test configuration and exit.\n"\
 					"        -v     : Display version and SVN revision of Corosync and exit.\n");
 				logsys_system_fini();
@@ -1319,7 +1274,8 @@ int main (int argc, char **argv, char **envp)
 #endif
 
 	if (icmap_init() != CS_OK) {
-		log_printf (LOGSYS_LEVEL_ERROR, "Corosync Executive couldn't initialize configuration component.");
+		fprintf (stderr, "Corosync Executive couldn't initialize configuration component.\n");
+		syslog (LOGSYS_LEVEL_ERROR, "Corosync Executive couldn't initialize configuration component.");
 		corosync_exit_error (COROSYNC_DONE_ICMAP);
 	}
 	set_icmap_ro_keys_flag();
@@ -1340,6 +1296,12 @@ int main (int argc, char **argv, char **envp)
 		corosync_exit_error (COROSYNC_DONE_MAINCONFIGREAD);
 	}
 
+	if (stats_map_init(api) != CS_OK) {
+		fprintf (stderr, "Corosync Executive couldn't initialize statistics component.\n");
+		syslog (LOGSYS_LEVEL_ERROR, "Corosync Executive couldn't initialize statistics component.");
+		corosync_exit_error (COROSYNC_DONE_STATS);
+	}
+
 	res = corosync_log_config_read (&error_string);
 	if (res == -1) {
 		/*
@@ -1356,23 +1318,30 @@ int main (int argc, char **argv, char **envp)
 	}
 
 	if (!testonly) {
-		log_printf (LOGSYS_LEVEL_NOTICE, "Corosync Cluster Engine ('%s'): started and ready to provide service.", VERSION);
+		log_printf (LOGSYS_LEVEL_NOTICE, "Corosync Cluster Engine %s starting up", VERSION);
 		log_printf (LOGSYS_LEVEL_INFO, "Corosync built-in features:" PACKAGE_FEATURES "");
 	}
 
+	/*
+	 * Create totem logsys subsys before totem_config_read so log functions can be used
+	 */
+	log_subsys_id_totem = _logsys_subsys_create("TOTEM", "totem,"
+			"totemip.c,totemconfig.c,totemcrypto.c,totemsrp.c,"
+			"totempg.c,totemudp.c,totemudpu.c,totemnet.c,totemknet.c");
+
 	/*
 	 * Make sure required directory is present
 	 */
-	res = stat (get_run_dir(), &stat_out);
+	res = stat (get_state_dir(), &stat_out);
 	if ((res == -1) || (res == 0 && !S_ISDIR(stat_out.st_mode))) {
-		log_printf (LOGSYS_LEVEL_ERROR, "Required directory not present %s.  Please create it.", get_run_dir());
+		log_printf (LOGSYS_LEVEL_ERROR, "State directory %s not present.  Please create it.", get_state_dir());
 		corosync_exit_error (COROSYNC_DONE_DIR_NOT_PRESENT);
 	}
 
-	res = chdir(get_run_dir());
+	res = chdir(get_state_dir());
 	if (res == -1) {
-		log_printf (LOGSYS_LEVEL_ERROR, "Cannot chdir to run directory %s.  "
-		    "Please make sure it has correct context and rights.", get_run_dir());
+		log_printf (LOGSYS_LEVEL_ERROR, "Cannot chdir to state directory %s.  "
+		    "Please make sure it has correct context and rights.", get_state_dir());
 		corosync_exit_error (COROSYNC_DONE_DIR_NOT_PRESENT);
 	}
 
@@ -1420,6 +1389,14 @@ int main (int argc, char **argv, char **envp)
 	}
 
 
+	move_to_root_cgroup = 1;
+	if (icmap_get_string("system.move_to_root_cgroup", &tmp_str) == CS_OK) {
+		if (strcmp(tmp_str, "yes") != 0) {
+			move_to_root_cgroup = 0;
+		}
+		free(tmp_str);
+	}
+
 	/*
 	 * Try to move corosync into root cpu cgroup. Failure is not fatal and
 	 * error is deliberately ignored.
@@ -1428,6 +1405,35 @@ int main (int argc, char **argv, char **envp)
 		(void)corosync_move_to_root_cgroup();
 	}
 
+	sched_rr = 1;
+	if (icmap_get_string("system.sched_rr", &tmp_str) == CS_OK) {
+		if (strcmp(tmp_str, "yes") != 0) {
+			sched_rr = 0;
+		}
+		free(tmp_str);
+	}
+
+	prio = 0;
+	if (icmap_get_string("system.priority", &tmp_str) == CS_OK) {
+		if (strcmp(tmp_str, "max") == 0) {
+			prio = INT_MIN;
+		} else if (strcmp(tmp_str, "min") == 0) {
+			prio = INT_MAX;
+		} else {
+			errno = 0;
+
+			tmpli = strtol(tmp_str, &ep, 10);
+			if (errno != 0 || *ep != '\0' || tmpli > INT_MAX || tmpli < INT_MIN) {
+				log_printf (LOGSYS_LEVEL_ERROR, "Priority value %s is invalid", tmp_str);
+				corosync_exit_error (COROSYNC_DONE_MAINCONFIGREAD);
+			}
+
+			prio = tmpli;
+		}
+
+		free(tmp_str);
+	}
+
 	/*
 	 * Set round robin realtime scheduling with priority 99
 	 */
@@ -1446,15 +1452,11 @@ int main (int argc, char **argv, char **envp)
 		}
 	}
 
-	ip_version = totem_config.ip_version;
-
 	totem_config.totem_memb_ring_id_create_or_load = corosync_ring_id_create_or_load;
 	totem_config.totem_memb_ring_id_store = corosync_ring_id_store;
 
 	totem_config.totem_logging_configuration = totem_logging_configuration;
-	totem_config.totem_logging_configuration.log_subsys_id = _logsys_subsys_create("TOTEM", "totem,"
-			"totemmrp.c,totemrrp.c,totemip.c,totemconfig.c,totemcrypto.c,totemsrp.c,"
-			"totempg.c,totemiba.c,totemudp.c,totemudpu.c,totemnet.c");
+	totem_config.totem_logging_configuration.log_subsys_id = log_subsys_id_totem;
 
 	totem_config.totem_logging_configuration.log_level_security = LOGSYS_LEVEL_WARNING;
 	totem_config.totem_logging_configuration.log_level_error = LOGSYS_LEVEL_ERROR;
@@ -1463,6 +1465,7 @@ int main (int argc, char **argv, char **envp)
 	totem_config.totem_logging_configuration.log_level_debug = LOGSYS_LEVEL_DEBUG;
 	totem_config.totem_logging_configuration.log_level_trace = LOGSYS_LEVEL_TRACE;
 	totem_config.totem_logging_configuration.log_printf = _logsys_log_printf;
+
 	logsys_config_apply();
 
 	/*
