@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2005-2011 Red Hat, Inc.
+ * Copyright (c) 2005-2019 Red Hat, Inc.
  *
  * All rights reserved.
  *
@@ -50,9 +50,9 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <ifaddrs.h>
-#include <corosync/logsys.h>
 
 #include <corosync/totem/totemip.h>
+#include <corosync/logsys.h>
 #include <corosync/swab.h>
 
 #define LOCALHOST_IPV4 "127.0.0.1"
@@ -92,33 +92,6 @@ int totemip_equal(const struct totem_ip_address *addr1,
 
 }
 
-int totemip_sa_equal(const struct totem_ip_address *totem_ip,
-	const struct sockaddr *sa)
-{
-	int res;
-
-	res = 0;
-
-	if (totem_ip->family != sa->sa_family) {
-		return (res);
-	}
-
-	switch (totem_ip->family) {
-	case AF_INET:
-		res = (memcmp(totem_ip->addr,
-		    &((const struct sockaddr_in *)sa)->sin_addr, sizeof(struct in_addr)) == 0);
-		break;
-	case AF_INET6:
-		res =  (memcmp(totem_ip->addr,
-		    &((const struct sockaddr_in6 *)sa)->sin6_addr, sizeof(struct in6_addr)) == 0);
-		break;
-	default:
-		assert(0);
-	}
-
-	return (res);
-}
-
 /* Copy a totem_ip_address */
 void totemip_copy(struct totem_ip_address *addr1,
 		  const struct totem_ip_address *addr2)
@@ -268,58 +241,11 @@ const char *totemip_print(const struct totem_ip_address *addr)
 	return (inet_ntop(addr->family, addr->addr, buf, sizeof(buf)));
 }
 
-static int totemip_getif_scopeid(const unsigned char *addr16, unsigned int *scopeid)
-{
-	struct ifaddrs *ifa;
-	const struct sockaddr_in6 *sin6;
-	const socklen_t addr_len = sizeof(struct in6_addr);
-	int rc = -1; // 0 = found 1 match; -1 = found 0 matches; -2 = found >1 matches
-	struct ifaddrs *totemip_getif_scopeid_ifap;
-
-	if (getifaddrs(&totemip_getif_scopeid_ifap) != 0) {
-		return (-1);
-	}
-
-	for (ifa = totemip_getif_scopeid_ifap; ifa; ifa = ifa->ifa_next) {
-		if (ifa->ifa_addr == NULL || ifa->ifa_netmask == NULL)
-			continue ;
-
-		if ((ifa->ifa_addr->sa_family != AF_INET6) ||
-		    (ifa->ifa_netmask->sa_family != AF_INET6 &&
-		     ifa->ifa_netmask->sa_family != 0))
-			continue ;
-
-		sin6 = (const struct sockaddr_in6 *)ifa->ifa_addr;
-
-		if (memcmp(&sin6->sin6_addr, addr16, addr_len) == 0) {
-			*scopeid = sin6->sin6_scope_id;
-
-			if (rc == -1) {
-				rc = 0;
-			} else {
-				rc = -2;
-			}
-		}
-	}
-
-	if (rc == -2) {
-		log_printf(LOGSYS_LEVEL_WARNING, "(%s) exists on several interfaces."
-		  " Use 'ip a' to see details.",
-		    totemip_sa_print(ifa->ifa_addr));
-	}
-
-	freeifaddrs(totemip_getif_scopeid_ifap);
-
-	return rc;
-}
-
-
-int totemip_totemip_to_sockaddr_convert_with_scopeid(const struct totem_ip_address *ip_addr,
-					uint16_t port, struct sockaddr_storage *saddr, int *addrlen,
-					int fill_scopeid)
+/* Make a totem_ip_address into a usable sockaddr_storage */
+int totemip_totemip_to_sockaddr_convert(struct totem_ip_address *ip_addr,
+					uint16_t port, struct sockaddr_storage *saddr, int *addrlen)
 {
 	int ret = -1;
-	unsigned int scopeid;
 
 	if (ip_addr->family == AF_INET) {
 		struct sockaddr_in *sin = (struct sockaddr_in *)saddr;
@@ -344,11 +270,7 @@ int totemip_totemip_to_sockaddr_convert_with_scopeid(const struct totem_ip_addre
 #endif
 		sin->sin6_family = ip_addr->family;
 		sin->sin6_port = ntohs(port);
-		if (fill_scopeid) {
-			if (totemip_getif_scopeid(ip_addr->addr, &scopeid) == 0) {
-				sin->sin6_scope_id = scopeid;
-			}
-		}
+		sin->sin6_scope_id = 2;
 		memcpy(&sin->sin6_addr, ip_addr->addr, sizeof(struct in6_addr));
 
 		*addrlen = sizeof(struct sockaddr_in6);
@@ -358,47 +280,121 @@ int totemip_totemip_to_sockaddr_convert_with_scopeid(const struct totem_ip_addre
 	return ret;
 }
 
-/* Make a totem_ip_address into a usable sockaddr_storage */
-int totemip_totemip_to_sockaddr_convert(struct totem_ip_address *ip_addr,
-					uint16_t port, struct sockaddr_storage *saddr, int *addrlen)
-{
-
-	return (totemip_totemip_to_sockaddr_convert_with_scopeid(ip_addr, port, saddr, addrlen, 0));
-}
-
-
-/* Converts an address string string into a totem_ip_address.
-   family can be AF_INET, AF_INET6 or 0 ("for "don't care")
-*/
-int totemip_parse(struct totem_ip_address *totemip, const char *addr, int family)
+/*
+ * Converts an address string string into a totem_ip_address. ip_version enum
+ * defines order.
+ */
+int totemip_parse(struct totem_ip_address *totemip, const char *addr,
+    enum totem_ip_version_enum ip_version)
 {
 	struct addrinfo *ainfo;
+	struct addrinfo *ainfo_iter;
+	struct addrinfo *ainfo_ipv4;
+	struct addrinfo *ainfo_ipv6;
+	struct addrinfo *ainfo_final;
 	struct addrinfo ahints;
 	struct sockaddr_in *sa;
 	struct sockaddr_in6 *sa6;
 	int ret;
+	int debug_ip_family;
+	int ai_family;
 
 	memset(&ahints, 0, sizeof(ahints));
 	ahints.ai_socktype = SOCK_DGRAM;
 	ahints.ai_protocol = IPPROTO_UDP;
-	ahints.ai_family   = family;
 
-	/* Lookup the nodename address */
+	ai_family = AF_UNSPEC;
+	debug_ip_family = 0;
+
+	switch (ip_version) {
+	case TOTEM_IP_VERSION_4:
+		ai_family = AF_INET;
+		debug_ip_family = 4;
+		break;
+	case TOTEM_IP_VERSION_6:
+		ai_family = AF_INET6;
+		debug_ip_family = 6;
+		break;
+	case TOTEM_IP_VERSION_6_4:
+	case TOTEM_IP_VERSION_4_6:
+		/*
+		 * ai_family and debug_ip_family are already set correctly
+		 */
+		break;
+	}
+
+	ahints.ai_family = ai_family;
+
 	ret = getaddrinfo(addr, NULL, &ahints, &ainfo);
-	if (ret)
-		return -1;
 
-	sa = (struct sockaddr_in *)ainfo->ai_addr;
-	sa6 = (struct sockaddr_in6 *)ainfo->ai_addr;
-	totemip->family = ainfo->ai_family;
+	if (ret == 0 && ai_family == AF_UNSPEC) {
+		ainfo_ipv4 = ainfo_ipv6 = NULL;
 
-	if (ainfo->ai_family == AF_INET)
+		/*
+		 * Walk thru results and store first AF_INET and AF_INET6
+		 */
+		for (ainfo_iter = ainfo; ainfo_iter != NULL; ainfo_iter = ainfo_iter->ai_next) {
+			if (ainfo_iter->ai_family == AF_INET && ainfo_ipv4 == NULL) {
+				ainfo_ipv4 = ainfo_iter;
+			}
+
+			if (ainfo_iter->ai_family == AF_INET6 && ainfo_ipv6 == NULL) {
+				ainfo_ipv6 = ainfo_iter;
+			}
+		}
+
+		if (ip_version == TOTEM_IP_VERSION_6_4) {
+			if (ainfo_ipv6 != NULL) {
+				ainfo_final = ainfo_ipv6;
+			} else {
+				ainfo_final = ainfo_ipv4;
+			}
+		} else {
+			if (ainfo_ipv4 != NULL) {
+				ainfo_final = ainfo_ipv4;
+			} else {
+				ainfo_final = ainfo_ipv6;
+			}
+		}
+	} else if (ret == 0) {
+		ainfo_final = ainfo;
+	} else {
+		ainfo_final = NULL;
+	}
+
+	if (ainfo_final == NULL) {
+		if (ret == 0) {
+			freeaddrinfo(ainfo);
+		}
+
+		if (debug_ip_family == 0) {
+			log_printf(LOGSYS_LEVEL_DEBUG, "totemip_parse: IP address of %s not resolvable",
+			    addr);
+		} else {
+			log_printf(LOGSYS_LEVEL_DEBUG, "totemip_parse: IPv%u address of %s not resolvable",
+			    debug_ip_family, addr);
+		}
+
+		return (-1);
+	}
+
+	totemip->family = ainfo_final->ai_family;
+	if (ainfo_final->ai_family == AF_INET) {
+		sa = (struct sockaddr_in *)ainfo_final->ai_addr;
 		memcpy(totemip->addr, &sa->sin_addr, sizeof(struct in_addr));
-	else
+		debug_ip_family = 4;
+	} else {
+		sa6 = (struct sockaddr_in6 *)ainfo_final->ai_addr;
 		memcpy(totemip->addr, &sa6->sin6_addr, sizeof(struct in6_addr));
+		debug_ip_family = 6;
+	}
+
+	log_printf(LOGSYS_LEVEL_DEBUG, "totemip_parse: IPv%u address of %s resolved as %s",
+		    debug_ip_family, addr, totemip_print(totemip));
 
 	freeaddrinfo(ainfo);
-	return 0;
+
+	return (0);
 }
 
 /* Make a sockaddr_* into a totem_ip_address */
@@ -428,7 +424,7 @@ int totemip_sockaddr_to_totemip_convert(const struct sockaddr_storage *saddr,
 	return ret;
 }
 
-int totemip_getifaddrs(struct list_head *addrs)
+int totemip_getifaddrs(struct qb_list_head *addrs)
 {
 	struct ifaddrs *ifap, *ifa;
 	struct totem_ip_if_address *if_addr;
@@ -436,7 +432,7 @@ int totemip_getifaddrs(struct list_head *addrs)
 	if (getifaddrs(&ifap) != 0)
 		return (-1);
 
-	list_init(addrs);
+	qb_list_init(addrs);
 
 	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
 		if (ifa->ifa_addr == NULL || ifa->ifa_netmask == NULL)
@@ -456,7 +452,7 @@ int totemip_getifaddrs(struct list_head *addrs)
 			goto error_free_ifaddrs;
 		}
 
-		list_init(&if_addr->list);
+		qb_list_init(&if_addr->list);
 
 		memset(if_addr, 0, sizeof(struct totem_ip_if_address));
 
@@ -477,7 +473,7 @@ int totemip_getifaddrs(struct list_head *addrs)
 			goto error_free_addr_name;
 		}
 
-		list_add_tail(&if_addr->list, addrs);
+		qb_list_add_tail(&if_addr->list, addrs);
 	}
 
 	freeifaddrs(ifap);
@@ -496,20 +492,19 @@ error_free_ifaddrs:
 	return (-1);
 }
 
-void totemip_freeifaddrs(struct list_head *addrs)
+void totemip_freeifaddrs(struct qb_list_head *addrs)
 {
 	struct totem_ip_if_address *if_addr;
-	struct list_head *list;
+	struct qb_list_head *list, *tmp_iter;
 
-	for (list = addrs->next; list != addrs;) {
-		if_addr = list_entry(list, struct totem_ip_if_address, list);
-		list = list->next;
+	qb_list_for_each_safe(list, tmp_iter, addrs) {
+		if_addr = qb_list_entry(list, struct totem_ip_if_address, list);
 
 		free(if_addr->name);
-		list_del(&if_addr->list);
+		qb_list_del(&if_addr->list);
 	        free(if_addr);
 	}
-	list_init(addrs);
+	qb_list_init(addrs);
 }
 
 int totemip_iface_check(struct totem_ip_address *bindnet,
@@ -518,8 +513,8 @@ int totemip_iface_check(struct totem_ip_address *bindnet,
 			int *interface_num,
 			int mask_high_bit)
 {
-	struct list_head addrs;
-	struct list_head *list;
+	struct qb_list_head addrs;
+	struct qb_list_head *list;
 	struct totem_ip_if_address *if_addr;
 	struct totem_ip_address bn_netaddr, if_netaddr;
 	socklen_t addr_len;
@@ -535,8 +530,8 @@ int totemip_iface_check(struct totem_ip_address *bindnet,
 		return (-1);
 	}
 
-	for (list = addrs.next; list != &addrs; list = list->next) {
-		if_addr = list_entry(list, struct totem_ip_if_address, list);
+	qb_list_for_each(list, &addrs) {
+		if_addr = qb_list_entry(list, struct totem_ip_if_address, list);
 
 		if (bindnet->family != if_addr->ip_addr.family)
 			continue ;
@@ -573,18 +568,6 @@ int totemip_iface_check(struct totem_ip_address *bindnet,
 			*interface_up = if_addr->interface_up;
 			*interface_num = if_addr->interface_num;
 
-			if (boundto->family == AF_INET && boundto->nodeid == 0) {
-				unsigned int nodeid = 0;
-				memcpy (&nodeid, boundto->addr, sizeof (int));
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-                                nodeid = swab32 (nodeid);
-#endif
-				if (mask_high_bit) {
-					nodeid &= 0x7FFFFFFF;
-				}
-				boundto->nodeid = nodeid;
-			}
-
 			net_match_found = 1;
 			res = 0;
 
