@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009-2020 Red Hat, Inc.
+ * Copyright (c) 2009-2015 Red Hat, Inc.
  *
  * All rights reserved.
  *
@@ -41,11 +41,11 @@
 #include <stdint.h>
 #include <unistd.h>
 
+#include <qb/qblist.h>
 #include <qb/qbipc_common.h>
 
 #include "quorum.h"
 #include <corosync/corodefs.h>
-#include <corosync/list.h>
 #include <corosync/logsys.h>
 #include <corosync/coroapi.h>
 #include <corosync/icmap.h>
@@ -80,7 +80,6 @@ static uint8_t two_node = 0;
 
 static uint8_t wait_for_all = 0;
 static uint8_t wait_for_all_status = 0;
-static uint8_t wait_for_all_autoset = 0; /* Wait for all is not set explicitly and follows two_node */
 
 static enum {ATB_NONE, ATB_LOWEST, ATB_HIGHEST, ATB_LIST} auto_tie_breaker = ATB_NONE, initial_auto_tie_breaker = ATB_NONE;
 static int lowest_node_id = -1;
@@ -187,7 +186,7 @@ struct cluster_node {
 	uint32_t    votes;
 	uint32_t    expected_votes;
 	uint32_t    flags;
-	struct      list_head list;
+	struct      qb_list_head list;
 };
 
 /*
@@ -202,7 +201,7 @@ static uint8_t cluster_is_quorate;
  */
 
 static struct cluster_node *us;
-static struct list_head cluster_members_list;
+static struct qb_list_head cluster_members_list;
 static unsigned int quorum_members[PROCESSOR_COUNT_MAX];
 static unsigned int previous_quorum_members[PROCESSOR_COUNT_MAX];
 static unsigned int atb_nodelist[PROCESSOR_COUNT_MAX];
@@ -224,11 +223,11 @@ struct quorum_pd {
 	unsigned char track_flags;
 	int tracking_enabled;
 	uint64_t tracking_context;
-	struct list_head list;
+	struct qb_list_head list;
 	void *conn;
 };
 
-static struct list_head trackers_list;
+static struct qb_list_head trackers_list;
 
 /*
  * votequorum timers
@@ -432,31 +431,24 @@ static struct default_service votequorum_service[] = {
 
 #define max(a,b) (((a) > (b)) ? (a) : (b))
 
-#define list_iterate(v, head) \
-	for (v = (head)->next; v != head; v = v->next)
-
 static void node_add_ordered(struct cluster_node *newnode)
 {
 	struct cluster_node *node = NULL;
-	struct list_head *tmp;
-	struct list_head *newlist = &newnode->list;
+	struct qb_list_head *tmp;
 
 	ENTER();
 
-	list_iterate(tmp, &cluster_members_list) {
-		node = list_entry(tmp, struct cluster_node, list);
+	qb_list_for_each(tmp, &cluster_members_list) {
+		node = qb_list_entry(tmp, struct cluster_node, list);
 		if (newnode->node_id < node->node_id) {
 			break;
 		}
 	}
 
 	if (!node) {
-		list_add(&newnode->list, &cluster_members_list);
+		qb_list_add(&newnode->list, &cluster_members_list);
 	} else {
-		newlist->prev = tmp->prev;
-		newlist->next = tmp;
-		tmp->prev->next = newlist;
-		tmp->prev = newlist;
+		qb_list_add_tail(&newnode->list, &node->list);
 	}
 
 	LEAVE();
@@ -465,7 +457,7 @@ static void node_add_ordered(struct cluster_node *newnode)
 static struct cluster_node *allocate_node(unsigned int nodeid)
 {
 	struct cluster_node *cl = NULL;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 
 	ENTER();
 
@@ -473,8 +465,8 @@ static struct cluster_node *allocate_node(unsigned int nodeid)
 		cl = (struct cluster_node *)&cluster_nodes[cluster_nodes_entries];
 		cluster_nodes_entries++;
 	} else {
-		list_iterate(tmp, &cluster_members_list) {
-			cl = list_entry(tmp, struct cluster_node, list);
+		qb_list_for_each(tmp, &cluster_members_list) {
+			cl = qb_list_entry(tmp, struct cluster_node, list);
 			if (cl->state == NODESTATE_DEAD) {
 				break;
 			}
@@ -486,7 +478,7 @@ static struct cluster_node *allocate_node(unsigned int nodeid)
 			log_printf(LOGSYS_LEVEL_CRIT, "Unable to find memory for node %u data!!", nodeid);
 			goto out;
 		}
-		list_del(tmp);
+		qb_list_del(tmp);
 	}
 
 	memset(cl, 0, sizeof(struct cluster_node));
@@ -504,7 +496,7 @@ out:
 static struct cluster_node *find_node_by_nodeid(unsigned int nodeid)
 {
 	struct cluster_node *node;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 
 	ENTER();
 
@@ -518,8 +510,8 @@ static struct cluster_node *find_node_by_nodeid(unsigned int nodeid)
 		return qdevice;
 	}
 
-	list_iterate(tmp, &cluster_members_list) {
-		node = list_entry(tmp, struct cluster_node, list);
+	qb_list_for_each(tmp, &cluster_members_list) {
+		node = qb_list_entry(tmp, struct cluster_node, list);
 		if (node->node_id == nodeid) {
 			LEAVE();
 			return node;
@@ -533,14 +525,14 @@ static struct cluster_node *find_node_by_nodeid(unsigned int nodeid)
 static void get_lowest_node_id(void)
 {
 	struct cluster_node *node = NULL;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 
 	ENTER();
 
 	lowest_node_id = us->node_id;
 
-	list_iterate(tmp, &cluster_members_list) {
-		node = list_entry(tmp, struct cluster_node, list);
+	qb_list_for_each(tmp, &cluster_members_list) {
+		node = qb_list_entry(tmp, struct cluster_node, list);
 		if ((node->state == NODESTATE_MEMBER) &&
 		    (node->node_id < lowest_node_id)) {
 			lowest_node_id = node->node_id;
@@ -555,14 +547,14 @@ static void get_lowest_node_id(void)
 static void get_highest_node_id(void)
 {
 	struct cluster_node *node = NULL;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 
 	ENTER();
 
 	highest_node_id = us->node_id;
 
-	list_iterate(tmp, &cluster_members_list) {
-		node = list_entry(tmp, struct cluster_node, list);
+	qb_list_for_each(tmp, &cluster_members_list) {
+		node = qb_list_entry(tmp, struct cluster_node, list);
 		if ((node->state == NODESTATE_MEMBER) &&
 		    (node->node_id > highest_node_id)) {
 			highest_node_id = node->node_id;
@@ -577,13 +569,13 @@ static void get_highest_node_id(void)
 static int check_low_node_id_partition(void)
 {
 	struct cluster_node *node = NULL;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 	int found = 0;
 
 	ENTER();
 
-	list_iterate(tmp, &cluster_members_list) {
-		node = list_entry(tmp, struct cluster_node, list);
+	qb_list_for_each(tmp, &cluster_members_list) {
+		node = qb_list_entry(tmp, struct cluster_node, list);
 		if ((node->state == NODESTATE_MEMBER) &&
 		    (node->node_id == lowest_node_id)) {
 				found = 1;
@@ -597,13 +589,13 @@ static int check_low_node_id_partition(void)
 static int check_high_node_id_partition(void)
 {
 	struct cluster_node *node = NULL;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 	int found = 0;
 
 	ENTER();
 
-	list_iterate(tmp, &cluster_members_list) {
-		node = list_entry(tmp, struct cluster_node, list);
+	qb_list_for_each(tmp, &cluster_members_list) {
+		node = qb_list_entry(tmp, struct cluster_node, list);
 		if ((node->state == NODESTATE_MEMBER) &&
 		    (node->node_id == highest_node_id)) {
 				found = 1;
@@ -742,13 +734,13 @@ static void parse_atb_string(char *atb_string)
 static int check_qdevice_master(void)
 {
 	struct cluster_node *node = NULL;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 	int found = 0;
 
 	ENTER();
 
-	list_iterate(tmp, &cluster_members_list) {
-		node = list_entry(tmp, struct cluster_node, list);
+	qb_list_for_each(tmp, &cluster_members_list) {
+		node = qb_list_entry(tmp, struct cluster_node, list);
 		if ((node->state == NODESTATE_MEMBER) &&
 		    (node->flags & NODE_FLAGS_QDEVICE_MASTER_WINS) &&
 		    (node->flags & NODE_FLAGS_QDEVICE_CAST_VOTE)) {
@@ -788,7 +780,7 @@ static int load_ev_tracking_barrier(void)
 
 	ENTER();
 
-	snprintf(filename, sizeof(filename) - 1, "%s/ev_tracking", get_run_dir());
+	snprintf(filename, sizeof(filename) - 1, "%s/ev_tracking", get_state_dir());
 
 	ev_tracking_fd = open(filename, O_RDWR, 0700);
 	if (ev_tracking_fd != -1) {
@@ -912,7 +904,7 @@ static void update_ev_tracking_barrier(uint32_t ev_t_barrier)
 
 static int calculate_quorum(int allow_decrease, unsigned int max_expected, unsigned int *ret_total_votes)
 {
-	struct list_head *nodelist;
+	struct qb_list_head *nodelist;
 	struct cluster_node *node;
 	unsigned int total_votes = 0;
 	unsigned int highest_expected = 0;
@@ -925,8 +917,8 @@ static int calculate_quorum(int allow_decrease, unsigned int max_expected, unsig
 		max_expected = max(ev_barrier, max_expected);
 	}
 
-	list_iterate(nodelist, &cluster_members_list) {
-		node = list_entry(nodelist, struct cluster_node, list);
+	qb_list_for_each(nodelist, &cluster_members_list) {
+		node = qb_list_entry(nodelist, struct cluster_node, list);
 
 		log_printf(LOGSYS_LEVEL_DEBUG, "node %u state=%d, votes=%u, expected=%u",
 			   node->node_id, node->state, node->votes, node->expected_votes);
@@ -985,12 +977,12 @@ static int calculate_quorum(int allow_decrease, unsigned int max_expected, unsig
 
 static void update_node_expected_votes(int new_expected_votes)
 {
-	struct list_head *nodelist;
+	struct qb_list_head *nodelist;
 	struct cluster_node *node;
 
 	if (new_expected_votes) {
-		list_iterate(nodelist, &cluster_members_list) {
-			node = list_entry(nodelist, struct cluster_node, list);
+		qb_list_for_each(nodelist, &cluster_members_list) {
+			node = qb_list_entry(nodelist, struct cluster_node, list);
 
 			if (node->state == NODESTATE_MEMBER) {
 				node->expected_votes = new_expected_votes;
@@ -1016,7 +1008,7 @@ static void are_we_quorate(unsigned int total_votes)
 				   "Waiting for all cluster members. "
 				   "Current votes: %d expected_votes: %d",
 				   total_votes, us->expected_votes);
-			assert(!cluster_is_quorate);
+			cluster_is_quorate = 0;
 			return;
 		}
 		update_wait_for_all_status(0);
@@ -1084,13 +1076,13 @@ static void get_total_votes(unsigned int *totalvotes, unsigned int *current_memb
 {
 	unsigned int total_votes = 0;
 	unsigned int cluster_members = 0;
-	struct list_head *nodelist;
+	struct qb_list_head *nodelist;
 	struct cluster_node *node;
 
 	ENTER();
 
-	list_iterate(nodelist, &cluster_members_list) {
-		node = list_entry(nodelist, struct cluster_node, list);
+	qb_list_for_each(nodelist, &cluster_members_list) {
+		node = qb_list_entry(nodelist, struct cluster_node, list);
 		if (node->state == NODESTATE_MEMBER) {
 			cluster_members++;
 			total_votes += node->votes;
@@ -1157,7 +1149,7 @@ static int votequorum_read_nodelist_configuration(uint32_t *votes,
 	icmap_iter_t iter;
 	const char *iter_key;
 	char tmp_key[ICMAP_KEYNAME_MAXLEN];
-	uint32_t our_pos, node_pos;
+	uint32_t our_pos, node_pos, last_node_pos=-1;
 	uint32_t nodecount = 0;
 	uint32_t nodelist_expected_votes = 0;
 	uint32_t node_votes = 0;
@@ -1180,9 +1172,15 @@ static int votequorum_read_nodelist_configuration(uint32_t *votes,
 			continue;
 		}
 
-		if (strcmp(tmp_key, "ring0_addr") != 0) {
+		/*
+		 * If current node_pos is the same as the last_node_pos then skip it
+		 * so we only do the code below once per node.
+		 * (icmap keys are always in order)
+		 */
+		if (last_node_pos == node_pos) {
 			continue;
 		}
+		last_node_pos = node_pos;
 
 		nodecount++;
 
@@ -1273,10 +1271,10 @@ static char *votequorum_readconfig(int runtime)
 	/*
 	 * gather basic data here
 	 */
-	(void)icmap_get_uint32("quorum.expected_votes", &expected_votes);
+	icmap_get_uint32("quorum.expected_votes", &expected_votes);
 	have_nodelist = votequorum_read_nodelist_configuration(&node_votes, &node_count, &node_expected_votes);
 	have_qdevice = votequorum_qdevice_is_configured(&qdevice_votes);
-	(void)icmap_get_uint8("quorum.two_node", &two_node);
+	icmap_get_uint8("quorum.two_node", &two_node);
 
 	/*
 	 * do config verification and enablement
@@ -1317,15 +1315,17 @@ static char *votequorum_readconfig(int runtime)
 	 * Enable special features
 	 */
 	if (!runtime) {
-		(void)icmap_get_uint8("quorum.allow_downscale", &allow_downscale);
-		if (icmap_get_uint8("quorum.wait_for_all", &wait_for_all) != CS_OK) {
-			wait_for_all_autoset = 1;
+		if (two_node) {
+			wait_for_all = 1;
 		}
-		(void)icmap_get_uint8("quorum.last_man_standing", &last_man_standing);
-		(void)icmap_get_uint32("quorum.last_man_standing_window", &last_man_standing_window);
-		(void)icmap_get_uint8("quorum.expected_votes_tracking", &ev_tracking);
-		(void)icmap_get_uint8("quorum.auto_tie_breaker", &atb);
-		(void)icmap_get_string("quorum.auto_tie_breaker_node", &atb_string);
+
+		icmap_get_uint8("quorum.allow_downscale", &allow_downscale);
+		icmap_get_uint8("quorum.wait_for_all", &wait_for_all);
+		icmap_get_uint8("quorum.last_man_standing", &last_man_standing);
+		icmap_get_uint32("quorum.last_man_standing_window", &last_man_standing_window);
+		icmap_get_uint8("quorum.expected_votes_tracking", &ev_tracking);
+		icmap_get_uint8("quorum.auto_tie_breaker", &atb);
+		icmap_get_string("quorum.auto_tie_breaker_node", &atb_string);
 
 		/* auto_tie_breaker defaults to LOWEST */
 		if (atb) {
@@ -1361,15 +1361,6 @@ static char *votequorum_readconfig(int runtime)
 
 	}
 
-	/*
-	 * Changing of wait_for_all during runtime is not supported, but changing of two_node is
-	 * and two_node may set wfa if not configured explicitly. It is safe to unset it
-	 * (or set it back) when two_node changes.
-	 */
-	if (wait_for_all_autoset) {
-		wait_for_all = two_node;
-	}
-
 	/* two_node and auto_tie_breaker are not compatible as two_node uses
 	 * a fence race to decide quorum whereas ATB decides based on node id
 	 */
@@ -1526,7 +1517,7 @@ static char *votequorum_readconfig(int runtime)
 		us->expected_votes = node_expected_votes;
 	} else {
 		us->votes = 1;
-		(void)icmap_get_uint32("quorum.votes", &us->votes);
+		icmap_get_uint32("quorum.votes", &us->votes);
 	}
 
 	if (expected_votes) {
@@ -1548,15 +1539,7 @@ static char *votequorum_readconfig(int runtime)
 	update_ev_barrier(us->expected_votes);
 	update_two_node();
 	if (wait_for_all) {
-		if (!runtime) {
-			update_wait_for_all_status(1);
-		}
-	} else if (wait_for_all_autoset && wait_for_all_status) {
-		/*
-		 * Reset wait for all status for consistency when wfa is auto-unset by 2node.
-		 * wait_for_all_status would be ignored by are_we_quorate anyway.
-		 */
-		update_wait_for_all_status(0);
+		update_wait_for_all_status(1);
 	}
 
 out:
@@ -1585,12 +1568,14 @@ static void votequorum_refresh_config(
 		return ;
 	}
 
-	(void)icmap_get_uint8("quorum.cancel_wait_for_all", &cancel_wfa);
+	icmap_get_uint8("quorum.cancel_wait_for_all", &cancel_wfa);
 	if (strcmp(key_name, "quorum.cancel_wait_for_all") == 0 &&
 	    cancel_wfa >= 1) {
 	        icmap_set_uint8("quorum.cancel_wait_for_all", 0);
-		votequorum_exec_send_reconfigure(VOTEQUORUM_RECONFIG_PARAM_CANCEL_WFA,
-						 us->node_id, 0);
+		if (votequorum_exec_send_reconfigure(VOTEQUORUM_RECONFIG_PARAM_CANCEL_WFA,
+						     us->node_id, 0)) {
+			log_printf(LOGSYS_LEVEL_ERROR, "Failed to send Cancel WFA message to other nodes");
+		}
 		return;
 	}
 
@@ -1608,12 +1593,16 @@ static void votequorum_refresh_config(
 	votequorum_exec_send_nodeinfo(us->node_id);
 	votequorum_exec_send_nodeinfo(VOTEQUORUM_QDEVICE_NODEID);
 	if (us->votes != old_votes) {
-		votequorum_exec_send_reconfigure(VOTEQUORUM_RECONFIG_PARAM_NODE_VOTES,
-						 us->node_id, us->votes);
+		if (votequorum_exec_send_reconfigure(VOTEQUORUM_RECONFIG_PARAM_NODE_VOTES,
+						     us->node_id, us->votes)) {
+			log_printf(LOGSYS_LEVEL_ERROR, "Failed to send new votes message to other nodes");
+		}
 	}
 	if (us->expected_votes != old_expected_votes) {
-		votequorum_exec_send_reconfigure(VOTEQUORUM_RECONFIG_PARAM_EXPECTED_VOTES,
-						 us->node_id, us->expected_votes);
+		if (votequorum_exec_send_reconfigure(VOTEQUORUM_RECONFIG_PARAM_EXPECTED_VOTES,
+						     us->node_id, us->expected_votes)) {
+			log_printf(LOGSYS_LEVEL_ERROR, "Failed to send expected votes message to other nodes");
+		}
 	}
 
 	LEAVE();
@@ -1723,11 +1712,7 @@ static int votequorum_exec_send_qdevice_reconfigure(const char *oldname, const c
 
 	req_exec_quorum_qdevice_reconfigure.header.id = SERVICE_ID_MAKE(VOTEQUORUM_SERVICE, MESSAGE_REQ_EXEC_VOTEQUORUM_QDEVICE_RECONFIGURE);
 	req_exec_quorum_qdevice_reconfigure.header.size = sizeof(req_exec_quorum_qdevice_reconfigure);
-
-	assert(strlen(oldname) < sizeof(req_exec_quorum_qdevice_reconfigure.oldname));
 	strcpy(req_exec_quorum_qdevice_reconfigure.oldname, oldname);
-
-	assert(strlen(newname) < sizeof(req_exec_quorum_qdevice_reconfigure.newname));
 	strcpy(req_exec_quorum_qdevice_reconfigure.newname, newname);
 
 	iov[0].iov_base = (void *)&req_exec_quorum_qdevice_reconfigure;
@@ -1750,8 +1735,6 @@ static int votequorum_exec_send_qdevice_reg(uint32_t operation, const char *qdev
 	req_exec_quorum_qdevice_reg.header.id = SERVICE_ID_MAKE(VOTEQUORUM_SERVICE, MESSAGE_REQ_EXEC_VOTEQUORUM_QDEVICE_REG);
 	req_exec_quorum_qdevice_reg.header.size = sizeof(req_exec_quorum_qdevice_reg);
 	req_exec_quorum_qdevice_reg.operation = operation;
-
-	assert(strlen(qdevice_name_req) < sizeof(req_exec_quorum_qdevice_reg.qdevice_name));
 	strcpy(req_exec_quorum_qdevice_reg.qdevice_name, qdevice_name_req);
 
 	iov[0].iov_base = (void *)&req_exec_quorum_qdevice_reg;
@@ -1766,7 +1749,7 @@ static int votequorum_exec_send_qdevice_reg(uint32_t operation, const char *qdev
 static int votequorum_exec_send_quorum_notification(void *conn, uint64_t context)
 {
 	struct res_lib_votequorum_quorum_notification *res_lib_votequorum_notification;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 	struct cluster_node *node;
 	int i = 0;
 	int cluster_members = 0;
@@ -1777,8 +1760,8 @@ static int votequorum_exec_send_quorum_notification(void *conn, uint64_t context
 
 	log_printf(LOGSYS_LEVEL_DEBUG, "Sending quorum callback, quorate = %d", cluster_is_quorate);
 
-	list_iterate(tmp, &cluster_members_list) {
-		node = list_entry(tmp, struct cluster_node, list);
+	qb_list_for_each(tmp, &cluster_members_list) {
+		node = qb_list_entry(tmp, struct cluster_node, list);
 		cluster_members++;
         }
 	if (us->flags & NODE_FLAGS_QDEVICE_REGISTERED) {
@@ -1796,8 +1779,8 @@ static int votequorum_exec_send_quorum_notification(void *conn, uint64_t context
 	res_lib_votequorum_notification->header.error = CS_OK;
 
 	/* Send all known nodes and their states */
-	list_iterate(tmp, &cluster_members_list) {
-		node = list_entry(tmp, struct cluster_node, list);
+	qb_list_for_each(tmp, &cluster_members_list) {
+		node = qb_list_entry(tmp, struct cluster_node, list);
 		res_lib_votequorum_notification->node_list[i].nodeid = node->node_id;
 		res_lib_votequorum_notification->node_list[i++].state = node->state;
         }
@@ -1814,8 +1797,8 @@ static int votequorum_exec_send_quorum_notification(void *conn, uint64_t context
 	} else {
 		struct quorum_pd *qpd;
 
-		list_iterate(tmp, &trackers_list) {
-			qpd = list_entry(tmp, struct quorum_pd, list);
+		qb_list_for_each(tmp, &trackers_list) {
+			qpd = qb_list_entry(tmp, struct quorum_pd, list);
 			res_lib_votequorum_notification->context = qpd->tracking_context;
 			corosync_api->ipc_dispatch_send(qpd->conn, &buf, size);
 		}
@@ -1831,18 +1814,18 @@ static int votequorum_exec_send_nodelist_notification(void *conn, uint64_t conte
 	struct res_lib_votequorum_nodelist_notification *res_lib_votequorum_notification;
 	int i = 0;
 	int size;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 	char buf[sizeof(struct res_lib_votequorum_nodelist_notification) + sizeof(uint32_t) * quorum_members_entries];
 
 	ENTER();
 
-	log_printf(LOGSYS_LEVEL_DEBUG, "Sending nodelist callback. ring_id = %d/%lld", quorum_ringid.rep.nodeid, quorum_ringid.seq);
+	log_printf(LOGSYS_LEVEL_DEBUG, "Sending nodelist callback. ring_id = %d/%lld", quorum_ringid.nodeid, quorum_ringid.seq);
 
 	size = sizeof(struct res_lib_votequorum_nodelist_notification) + sizeof(uint32_t) * quorum_members_entries;
 
 	res_lib_votequorum_notification = (struct res_lib_votequorum_nodelist_notification *)&buf;
 	res_lib_votequorum_notification->node_list_entries = quorum_members_entries;
-	res_lib_votequorum_notification->ring_id.nodeid = quorum_ringid.rep.nodeid;
+	res_lib_votequorum_notification->ring_id.nodeid = quorum_ringid.nodeid;
 	res_lib_votequorum_notification->ring_id.seq = quorum_ringid.seq;
 	res_lib_votequorum_notification->context = context;
 
@@ -1862,8 +1845,8 @@ static int votequorum_exec_send_nodelist_notification(void *conn, uint64_t conte
 	} else {
 		struct quorum_pd *qpd;
 
-		list_iterate(tmp, &trackers_list) {
-			qpd = list_entry(tmp, struct quorum_pd, list);
+		qb_list_for_each(tmp, &trackers_list) {
+			qpd = qb_list_entry(tmp, struct quorum_pd, list);
 			res_lib_votequorum_notification->context = qpd->tracking_context;
 			corosync_api->ipc_dispatch_send(qpd->conn, &buf, size);
 		}
@@ -1878,7 +1861,7 @@ static void votequorum_exec_send_expectedvotes_notification(void)
 {
 	struct res_lib_votequorum_expectedvotes_notification res_lib_votequorum_expectedvotes_notification;
 	struct quorum_pd *qpd;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 
 	ENTER();
 
@@ -1889,8 +1872,8 @@ static void votequorum_exec_send_expectedvotes_notification(void)
 	res_lib_votequorum_expectedvotes_notification.header.error = CS_OK;
 	res_lib_votequorum_expectedvotes_notification.expected_votes = us->expected_votes;
 
-	list_iterate(tmp, &trackers_list) {
-		qpd = list_entry(tmp, struct quorum_pd, list);
+	qb_list_for_each(tmp, &trackers_list) {
+		qpd = qb_list_entry(tmp, struct quorum_pd, list);
 		res_lib_votequorum_expectedvotes_notification.context = qpd->tracking_context;
 		corosync_api->ipc_dispatch_send(qpd->conn, &res_lib_votequorum_expectedvotes_notification,
 						sizeof(struct res_lib_votequorum_expectedvotes_notification));
@@ -1952,7 +1935,7 @@ static void message_handler_req_exec_votequorum_qdevice_reg (
 	struct res_lib_votequorum_status res_lib_votequorum_status;
 	int wipe_qdevice_name = 1;
 	struct cluster_node *node = NULL;
-	struct list_head *tmp;
+	struct qb_list_head *tmp;
 	cs_error_t error = CS_OK;
 
 	ENTER();
@@ -2021,8 +2004,8 @@ static void message_handler_req_exec_votequorum_qdevice_reg (
 		qdevice_reg_conn = NULL;
 		break;
 	case VOTEQUORUM_QDEVICE_OPERATION_UNREGISTER:
-		list_iterate(tmp, &cluster_members_list) {
-			node = list_entry(tmp, struct cluster_node, list);
+		qb_list_for_each(tmp, &cluster_members_list) {
+			node = qb_list_entry(tmp, struct cluster_node, list);
 			if ((node->state == NODESTATE_MEMBER) &&
 			    (node->flags & NODE_FLAGS_QDEVICE_REGISTERED)) {
 				wipe_qdevice_name = 0;
@@ -2198,7 +2181,6 @@ static void message_handler_req_exec_votequorum_reconfigure (
 	{
 	case VOTEQUORUM_RECONFIG_PARAM_EXPECTED_VOTES:
 		update_node_expected_votes(req_exec_quorum_reconfigure->value);
-
 		votequorum_exec_send_expectedvotes_notification();
 		update_ev_barrier(req_exec_quorum_reconfigure->value);
 		if (ev_tracking) {
@@ -2274,8 +2256,8 @@ static char *votequorum_exec_init_fn (struct corosync_api_v1 *api)
 	/*
 	 * make sure we start clean
 	 */
-	list_init(&cluster_members_list);
-	list_init(&trackers_list);
+	qb_list_init(&cluster_members_list);
+	qb_list_init(&trackers_list);
 	qdevice = NULL;
 	us = NULL;
 	memset(cluster_nodes, 0, sizeof(cluster_nodes));
@@ -2505,7 +2487,7 @@ static int quorum_lib_init_fn (void *conn)
 
 	ENTER();
 
-	list_init (&pd->list);
+	qb_list_init (&pd->list);
 	pd->conn = conn;
 
 	LEAVE();
@@ -2519,8 +2501,8 @@ static int quorum_lib_exit_fn (void *conn)
 	ENTER();
 
 	if (quorum_pd->tracking_enabled) {
-		list_del (&quorum_pd->list);
-		list_init (&quorum_pd->list);
+		qb_list_del (&quorum_pd->list);
+		qb_list_init (&quorum_pd->list);
 	}
 
 	LEAVE();
@@ -2578,10 +2560,10 @@ static void message_handler_req_lib_votequorum_getinfo (void *conn, const void *
 	node = find_node_by_nodeid(nodeid);
 	if (node) {
 		struct cluster_node *iternode;
-		struct list_head *nodelist;
+		struct qb_list_head *nodelist;
 
-		list_iterate(nodelist, &cluster_members_list) {
-			iternode = list_entry(nodelist, struct cluster_node, list);
+		qb_list_for_each(nodelist, &cluster_members_list) {
+			iternode = qb_list_entry(nodelist, struct cluster_node, list);
 
 			if (iternode->state == NODESTATE_MEMBER) {
 				highest_expected =
@@ -2685,19 +2667,17 @@ static void message_handler_req_lib_votequorum_setexpected (void *conn, const vo
 	 */
 	newquorum = calculate_quorum(1, req_lib_votequorum_setexpected->expected_votes, &total_votes);
 	allow_downscale = allow_downscale_status;
-	/*
-	 * Setting expected_votes < total_votes doesn't make sense.
-	 * For quorate cluster prevent cluster to become unquorate.
-	 */
-	if (req_lib_votequorum_setexpected->expected_votes < total_votes ||
-	    (cluster_is_quorate && (newquorum > total_votes))) {
+	if (newquorum < total_votes / 2 ||
+	    newquorum > total_votes) {
 		error = CS_ERR_INVALID_PARAM;
 		goto error_exit;
 	}
 	update_node_expected_votes(req_lib_votequorum_setexpected->expected_votes);
 
-	votequorum_exec_send_reconfigure(VOTEQUORUM_RECONFIG_PARAM_EXPECTED_VOTES, us->node_id,
-					 req_lib_votequorum_setexpected->expected_votes);
+	if (votequorum_exec_send_reconfigure(VOTEQUORUM_RECONFIG_PARAM_EXPECTED_VOTES, us->node_id,
+					     req_lib_votequorum_setexpected->expected_votes)) {
+		error = CS_ERR_NO_RESOURCES;
+	}
 
 error_exit:
 	res_lib_votequorum_status.header.size = sizeof(res_lib_votequorum_status);
@@ -2743,8 +2723,10 @@ static void message_handler_req_lib_votequorum_setvotes (void *conn, const void
 		goto error_exit;
 	}
 
-	votequorum_exec_send_reconfigure(VOTEQUORUM_RECONFIG_PARAM_NODE_VOTES, nodeid,
-					 req_lib_votequorum_setvotes->votes);
+	if (votequorum_exec_send_reconfigure(VOTEQUORUM_RECONFIG_PARAM_NODE_VOTES, nodeid,
+					     req_lib_votequorum_setvotes->votes)) {
+		error = CS_ERR_NO_RESOURCES;
+	}
 
 error_exit:
 	res_lib_votequorum_status.header.size = sizeof(res_lib_votequorum_status);
@@ -2791,7 +2773,7 @@ static void message_handler_req_lib_votequorum_trackstart (void *conn,
 		quorum_pd->tracking_enabled = 1;
 		quorum_pd->tracking_context = req_lib_votequorum_trackstart->context;
 
-		list_add (&quorum_pd->list, &trackers_list);
+		qb_list_add (&quorum_pd->list, &trackers_list);
 	}
 
 response_send:
@@ -2815,8 +2797,8 @@ static void message_handler_req_lib_votequorum_trackstop (void *conn,
 	if (quorum_pd->tracking_enabled) {
 		error = CS_OK;
 		quorum_pd->tracking_enabled = 0;
-		list_del (&quorum_pd->list);
-		list_init (&quorum_pd->list);
+		qb_list_del (&quorum_pd->list);
+		qb_list_init (&quorum_pd->list);
 	} else {
 		error = CS_ERR_NOT_EXIST;
 	}
@@ -2969,12 +2951,12 @@ static void message_handler_req_lib_votequorum_qdevice_poll (void *conn,
 	}
 
 	if (us->flags & NODE_FLAGS_QDEVICE_REGISTERED) {
-		if (!(req_lib_votequorum_qdevice_poll->ring_id.nodeid == quorum_ringid.rep.nodeid &&
+		if (!(req_lib_votequorum_qdevice_poll->ring_id.nodeid == quorum_ringid.nodeid &&
 		      req_lib_votequorum_qdevice_poll->ring_id.seq == quorum_ringid.seq)) {
 			log_printf(LOGSYS_LEVEL_DEBUG, "Received poll ring id (%u.%"PRIu64") != last sync "
 			    "ring id (%u.%"PRIu64"). Ignoring poll call.",
 			    req_lib_votequorum_qdevice_poll->ring_id.nodeid, req_lib_votequorum_qdevice_poll->ring_id.seq,
-			    quorum_ringid.rep.nodeid, quorum_ringid.seq);
+			    quorum_ringid.nodeid, quorum_ringid.seq);
 			error = CS_ERR_MESSAGE_ERROR;
 			goto out;
 		}
