@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2011-2012 Red Hat, Inc.
+ * Copyright (c) 2011-2017 Red Hat, Inc.
  *
  * Author: Jan Friesse (jfriesse@redhat.com)
  *
@@ -315,7 +315,6 @@ extern cs_error_t icmap_get_int64_r(const icmap_map_t map, const char *key_name,
 extern cs_error_t icmap_get_uint64_r(const icmap_map_t map, const char *key_name, uint64_t *u64);
 extern cs_error_t icmap_get_float_r(const icmap_map_t map, const char *key_name, float *flt);
 extern cs_error_t icmap_get_double_r(const icmap_map_t map, const char *key_name, double *dbl);
-extern cs_error_t icmap_get_string_r(const icmap_map_t map, const char *key_name, char **str);
 
 /**
  * @brief Shortcut for icmap_get for string type.
@@ -543,6 +542,21 @@ extern void icmap_convert_name_to_valid_name(char *key_name);
  */
 extern cs_error_t icmap_copy_map(icmap_map_t dst_map, const icmap_map_t src_map);
 
+/*
+ * Returns length of value of given type, or 0 for string and binary data type
+ */
+size_t icmap_get_valuetype_len(icmap_value_types_t type);
+
+/*
+ * Converts track type of icmap to qb
+ */
+int32_t icmap_tt_to_qbtt(int32_t track_type);
+
+/*
+ * Convert track type of qb to icmap
+ */
+int32_t icmap_qbtt_to_tt(int32_t track_type);
+
 #ifdef __cplusplus
 }
 #endif
