@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2002-2005 MontaVista Software, Inc.
- * Copyright (c) 2006-2013 Red Hat, Inc.
+ * Copyright (c) 2006-2018 Red Hat, Inc.
  *
  * All rights reserved.
  *
@@ -367,38 +367,6 @@ exit_handle_put:
 	return (error);
 }
 
-cs_error_t
-corosync_cfg_ring_reenable (
-	corosync_cfg_handle_t cfg_handle)
-{
-	struct cfg_inst *cfg_inst;
-	struct req_lib_cfg_ringreenable req_lib_cfg_ringreenable;
-	struct res_lib_cfg_ringreenable res_lib_cfg_ringreenable;
-	cs_error_t error;
-	struct iovec iov;
-
-	error = hdb_error_to_cs(hdb_handle_get (&cfg_hdb, cfg_handle, (void *)&cfg_inst));
-	if (error != CS_OK) {
-		return (error);
-	}
-
-	req_lib_cfg_ringreenable.header.size = sizeof (struct req_lib_cfg_ringreenable);
-	req_lib_cfg_ringreenable.header.id = MESSAGE_REQ_CFG_RINGREENABLE;
-
-	iov.iov_base = (void *)&req_lib_cfg_ringreenable,
-	iov.iov_len = sizeof (struct req_lib_cfg_ringreenable);
-
-	error = qb_to_cs_error (qb_ipcc_sendv_recv (cfg_inst->c,
-		&iov,
-		1,
-		&res_lib_cfg_ringreenable,
-		sizeof (struct res_lib_cfg_ringreenable), CS_IPC_TIMEOUT_MS));
-
-	(void)hdb_handle_put (&cfg_hdb, cfg_handle);
-
-	return (error);
-}
-
 cs_error_t
 corosync_cfg_kill_node (
 	corosync_cfg_handle_t cfg_handle,
@@ -512,7 +480,7 @@ corosync_cfg_replyto_shutdown (
 
 cs_error_t corosync_cfg_get_node_addrs (
 	corosync_cfg_handle_t cfg_handle,
-	int nodeid,
+	unsigned int nodeid,
 	size_t max_addrs,
 	int *num_addrs,
 	corosync_cfg_node_address_t *addrs)
@@ -526,12 +494,14 @@ cs_error_t corosync_cfg_get_node_addrs (
 	struct iovec iov;
 	const char *addr_buf;
 	char response_buf[IPC_RESPONSE_SIZE];
+	char zeroes[sizeof(struct sockaddr_storage)];
 
 	error = hdb_error_to_cs(hdb_handle_get (&cfg_hdb, cfg_handle,
 		(void *)&cfg_inst));
 	if (error != CS_OK) {
 		return (error);
 	}
+	memset(zeroes, 0, sizeof(zeroes));
 
 	req_lib_cfg_get_node_addrs.header.size = sizeof (req_lib_cfg_get_node_addrs);
 	req_lib_cfg_get_node_addrs.header.id = MESSAGE_REQ_CFG_GET_NODE_ADDRS;
@@ -565,14 +535,26 @@ cs_error_t corosync_cfg_get_node_addrs (
 
 		if (res_lib_cfg_get_node_addrs->family == AF_INET) {
 			in = (struct sockaddr_in *)addrs[i].address;
-			in->sin_family = AF_INET;
+			if (memcmp(addr_buf, zeroes, addrlen) == 0) {
+				in->sin_family = 0;
+			} else {
+				in->sin_family = AF_INET;
+			}
 			memcpy(&in->sin_addr, addr_buf, sizeof(struct in_addr));
 		}
 		if (res_lib_cfg_get_node_addrs->family == AF_INET6) {
 			in6 = (struct sockaddr_in6 *)addrs[i].address;
-			in6->sin6_family = AF_INET6;
+
+			if (memcmp(addr_buf, zeroes, addrlen) == 0) {
+				in6->sin6_family = 0;
+			} else {
+				in6->sin6_family = AF_INET6;
+			}
 			memcpy(&in6->sin6_addr, addr_buf, sizeof(struct in6_addr));
 		}
+
+		/* Mark it as unused */
+
 	}
 	*num_addrs = res_lib_cfg_get_node_addrs->num_addrs;
 	errno = error = res_lib_cfg_get_node_addrs->header.error;
@@ -663,3 +645,42 @@ error_exit:
 
 	return (error);
 }
+
+cs_error_t corosync_cfg_reopen_log_files (
+	corosync_cfg_handle_t handle)
+{
+	cs_error_t error;
+	struct cfg_inst *cfg_inst;
+	struct iovec iov;
+	struct req_lib_cfg_reopen_log_files req_lib_cfg_reopen_log_files;
+	struct res_lib_cfg_reopen_log_files res_lib_cfg_reopen_log_files;
+
+	error = hdb_error_to_cs(hdb_handle_get (&cfg_hdb, handle, (void *)&cfg_inst));
+	if (error != CS_OK) {
+		return (error);
+	}
+
+	req_lib_cfg_reopen_log_files.header.size = sizeof (struct qb_ipc_request_header);
+	req_lib_cfg_reopen_log_files.header.id = MESSAGE_REQ_CFG_REOPEN_LOG_FILES;
+
+	iov.iov_base = (void *)&req_lib_cfg_reopen_log_files;
+	iov.iov_len = sizeof (struct req_lib_cfg_reopen_log_files);
+
+	error = qb_to_cs_error (qb_ipcc_sendv_recv (
+		cfg_inst->c,
+		&iov,
+		1,
+		&res_lib_cfg_reopen_log_files,
+		sizeof (struct res_lib_cfg_reopen_log_files), CS_IPC_TIMEOUT_MS));
+
+	if (error != CS_OK) {
+		goto error_exit;
+	}
+
+	error = res_lib_cfg_reopen_log_files.header.error;
+
+error_exit:
+	(void)hdb_handle_put (&cfg_hdb, handle);
+
+	return (error);
+}
