@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2013 Red Hat, Inc.
+ * Copyright (c) 2006-2017 Red Hat, Inc.
  *
  * All rights reserved.
  *
@@ -46,12 +46,10 @@
 #include <sys/un.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
-#include <limits.h>
 
 #include <corosync/corotypes.h>
 #include <corosync/totem/totem.h>
 #include <corosync/cfg.h>
-#include "util.h"
 
 #define cs_repeat(result, max, code)				\
 	do {							\
@@ -67,7 +65,18 @@
 		} while (counter < max);			\
 	} while (0)
 
-static int ringstatusget_do (char *interface_name)
+enum user_action {
+	ACTION_NOOP=0,
+	ACTION_LINKSTATUS_GET,
+	ACTION_RELOAD_CONFIG,
+	ACTION_REOPEN_LOG_FILES,
+	ACTION_SHUTDOW,
+	ACTION_SHOWADDR,
+	ACTION_KILL_NODE,
+};
+
+static int
+linkstatusget_do (char *interface_name, int brief)
 {
 	cs_error_t result;
 	corosync_cfg_handle_t handle;
@@ -76,18 +85,19 @@ static int ringstatusget_do (char *interface_name)
 	char **interface_status;
 	unsigned int i;
 	unsigned int nodeid;
-	int rc = EXIT_SUCCESS;
+	int rc = 0;
+	int len, s = 0, t;
 
-	printf ("Printing ring status.\n");
+	printf ("Printing link status.\n");
 	result = corosync_cfg_initialize (&handle, NULL);
 	if (result != CS_OK) {
-		fprintf (stderr, "Could not initialize corosync configuration API error %d\n", result);
-		exit (EXIT_FAILURE);
+		printf ("Could not initialize corosync configuration API error %d\n", result);
+		exit (1);
 	}
 
 	result = corosync_cfg_local_get(handle, &nodeid);
 	if (result != CS_OK) {
-		fprintf (stderr, "Could not get the local node id, the error is: %d\n", result);
+		printf ("Could not get the local node id, the error is: %d\n", result);
 	}
 	else {
 		printf ("Local node ID %u\n", nodeid);
@@ -98,19 +108,43 @@ static int ringstatusget_do (char *interface_name)
 				&interface_status,
 				&interface_count);
 	if (result != CS_OK) {
-		fprintf (stderr, "Could not get the ring status, the error is: %d\n", result);
+		printf ("Could not get the link status, the error is: %d\n", result);
 	} else {
 		for (i = 0; i < interface_count; i++) {
-			if ( (interface_name && 
-			     	(interface_name[0]=='\0' || 
+			s = 0;
+			if ( (interface_name &&
+			      interface_names[i][0] != '\0' &&
+			      (interface_name[0]=='\0' ||
 				strcasecmp (interface_name, interface_names[i]) == 0)) ||
 				!interface_name ) {
 
-				printf ("RING ID %d\n", i);
-				printf ("\tid\t= %s\n", interface_names[i]);
-				printf ("\tstatus\t= %s\n", interface_status[i]);
-				if (strstr(interface_status[i], "FAULTY")) {
-					rc = EXIT_FAILURE;
+				/*
+				 * Interface_name is "<linkid> <IP address>"
+				 * separate them out
+				 */
+				char *space = strchr(interface_names[i], ' ');
+				if (!space) {
+					continue;
+				}
+				*space = '\0';
+
+				printf ("LINK ID %s\n", interface_names[i]);
+				printf ("\taddr\t= %s\n", space+1);
+				if((!brief) && (strcmp(interface_status[i], "OK") != 0) &&
+					(!strstr(interface_status[i], "FAULTY"))) {
+					len = strlen(interface_status[i]);
+					printf ("\tstatus:\n");
+					while(s < len) {
+						t = interface_status[i][s] - '0';
+						printf("\t\tnode %d:\t", s++);
+						printf("link enabled:%d\t", t&1? 1 : 0);
+						printf("link connected:%d\n", t&2? 1: 0);
+					}
+				} else {
+					printf ("\tstatus\t= %s\n", interface_status[i]);
+					if (strstr(interface_status[i], "FAULTY")) {
+						rc = 1;
+					}
 				}
 			}
 		}
@@ -127,44 +161,24 @@ static int ringstatusget_do (char *interface_name)
 	return rc;
 }
 
-static void ringreenable_do (void)
-{
-	cs_error_t result;
-	corosync_cfg_handle_t handle;
-
-	printf ("Re-enabling all failed rings.\n");
-	result = corosync_cfg_initialize (&handle, NULL);
-	if (result != CS_OK) {
-		fprintf (stderr, "Could not initialize corosync configuration API error %d\n", result);
-		exit (EXIT_FAILURE);
-	}
-
-	result = corosync_cfg_ring_reenable (handle);
-	if (result != CS_OK) {
-		fprintf (stderr, "Could not re-enable ring error %d\n", result);
-	}
-
-	(void)corosync_cfg_finalize (handle);
-}
-
 static int reload_config_do (void)
 {
 	cs_error_t result;
 	corosync_cfg_handle_t handle;
 	int rc;
 
-	rc = EXIT_SUCCESS;
+	rc = 0;
 
 	printf ("Reloading corosync.conf...\n");
 	result = corosync_cfg_initialize (&handle, NULL);
 	if (result != CS_OK) {
-		fprintf (stderr, "Could not initialize corosync configuration API error %s\n", cs_strerror(result));
-		exit (EXIT_FAILURE);
+		printf ("Could not initialize corosync configuration API error %s\n", cs_strerror(result));
+		exit (1);
 	}
 
 	result = corosync_cfg_reload_config (handle);
 	if (result != CS_OK) {
-		fprintf (stderr, "Could not reload configuration. Error %s\n", cs_strerror(result));
+		printf ("Could not reload configuration. Error %s\n", cs_strerror(result));
 		rc = (int)result;
 	}
 	else {
@@ -176,6 +190,31 @@ static int reload_config_do (void)
 	return (rc);
 }
 
+static int reopen_log_files_do (void)
+{
+	cs_error_t result;
+	corosync_cfg_handle_t handle;
+	int rc;
+
+	rc = 0;
+
+	result = corosync_cfg_initialize (&handle, NULL);
+	if (result != CS_OK) {
+		fprintf (stderr, "Could not initialize corosync configuration API error %s\n", cs_strerror(result));
+		exit (1);
+	}
+
+	result = corosync_cfg_reopen_log_files (handle);
+	if (result != CS_OK) {
+		fprintf (stderr, "Could not reopen corosync logging files. Error %s\n", cs_strerror(result));
+		rc = (int)result;
+	}
+
+	(void)corosync_cfg_finalize (handle);
+
+	return (rc);
+}
+
 static void shutdown_do(void)
 {
 	cs_error_t result;
@@ -186,14 +225,14 @@ static void shutdown_do(void)
 
 	result = corosync_cfg_initialize (&handle, &callbacks);
 	if (result != CS_OK) {
-		fprintf (stderr, "Could not initialize corosync configuration API error %d\n", result);
-		exit (EXIT_FAILURE);
+		printf ("Could not initialize corosync configuration API error %d\n", result);
+		exit (1);
 	}
 
 	printf ("Shutting down corosync\n");
 	cs_repeat(result, 30, corosync_cfg_try_shutdown (handle, COROSYNC_CFG_SHUTDOWN_FLAG_REQUEST));
 	if (result != CS_OK) {
-		fprintf (stderr, "Could not shutdown (error = %d)\n", result);
+		printf ("Could not shutdown (error = %d)\n", result);
 	}
 
 	(void)corosync_cfg_finalize (handle);
@@ -203,15 +242,16 @@ static void showaddrs_do(unsigned int nodeid)
 {
 	cs_error_t result;
 	corosync_cfg_handle_t handle;
+	corosync_cfg_callbacks_t callbacks;
 	int numaddrs;
 	int i;
 	corosync_cfg_node_address_t addrs[INTERFACE_MAX];
 
 
-	result = corosync_cfg_initialize (&handle, NULL);
+	result = corosync_cfg_initialize (&handle, &callbacks);
 	if (result != CS_OK) {
-		fprintf (stderr, "Could not initialize corosync configuration API error %d\n", result);
-		exit (EXIT_FAILURE);
+		printf ("Could not initialize corosync configuration API error %d\n", result);
+		exit (1);
 	}
 
 	if (corosync_cfg_get_node_addrs(handle, nodeid, INTERFACE_MAX, &numaddrs, addrs) == CS_OK) {
@@ -222,10 +262,14 @@ static void showaddrs_do(unsigned int nodeid)
 			struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)addrs[i].address;
 			void *saddr;
 
-			if (ss->ss_family == AF_INET6)
+			if (!ss->ss_family) {
+				continue;
+			}
+			if (ss->ss_family == AF_INET6) {
 				saddr = &sin6->sin6_addr;
-			else
+			} else {
 				saddr = &sin->sin_addr;
+			}
 
 			inet_ntop(ss->ss_family, saddr, buf, sizeof(buf));
 			if (i != 0) {
@@ -249,12 +293,12 @@ static void killnode_do(unsigned int nodeid)
 	printf ("Killing node %d\n", nodeid);
 	result = corosync_cfg_initialize (&handle, NULL);
 	if (result != CS_OK) {
-		fprintf (stderr, "Could not initialize corosync configuration API error %d\n", result);
-		exit (EXIT_FAILURE);
+		printf ("Could not initialize corosync configuration API error %d\n", result);
+		exit (1);
 	}
 	result = corosync_cfg_kill_node (handle, nodeid, "Killed by corosync-cfgtool");
 	if (result != CS_OK) {
-		fprintf (stderr, "Could not kill node (error = %d)\n", result);
+		printf ("Could not kill node (error = %d)\n", result);
 	}
 	(void)corosync_cfg_finalize (handle);
 }
@@ -262,14 +306,14 @@ static void killnode_do(unsigned int nodeid)
 
 static void usage_do (void)
 {
-	printf ("corosync-cfgtool [-i <interface ip>] [-s] [-r] [-R] [-k nodeid] [-a nodeid] [-h] [-H]\n\n");
+	printf ("corosync-cfgtool [[-i <interface ip>] [-b] -s] [-R] [-L] [-k nodeid] [-a nodeid] [-h] [-H]\n\n");
 	printf ("A tool for displaying and configuring active parameters within corosync.\n");
 	printf ("options:\n");
-	printf ("\t-i\tFinds only information about the specified interface IP address.\n");
-	printf ("\t-s\tDisplays the status of the current rings on this node.\n");
-	printf ("\t-r\tReset redundant ring state cluster wide after a fault to\n");
-	printf ("\t\tre-enable redundant ring operation.\n");
+	printf ("\t-i\tFinds only information about the specified interface IP address when used with -s..\n");
+	printf ("\t-s\tDisplays the status of the current links on this node(UDP/UDPU), while extended status for KNET.\n");
+	printf ("\t-b\tDisplays the brief status of the current links on this node when used with -s.(KNET only)\n");
 	printf ("\t-R\tTell all instances of corosync in this cluster to reload corosync.conf.\n");
+	printf ("\t-L\tTell corosync to reopen all logging files.\n");
 	printf ("\t-k\tKill a node identified by node id.\n");
 	printf ("\t-a\tDisplay the IP address(es) of a node\n");
 	printf ("\t-h\tPrint basic usage.\n");
@@ -277,16 +321,14 @@ static void usage_do (void)
 }
 
 int main (int argc, char *argv[]) {
-	const char *options = "i:srRk:a:hH";
+	const char *options = "i:sbrRLk:a:hH";
 	int opt;
-	unsigned int nodeid;
+	unsigned int nodeid = 0;
 	char interface_name[128] = "";
-	int rc = EXIT_SUCCESS;
-	long long int l;
+	int rc = 0;
+	enum user_action action = ACTION_NOOP;
+	int brief = 0;
 
-	if (argc == 1) {
-		usage_do ();
-	}
 	while ( (opt = getopt(argc, argv, options)) != -1 ) {
 		switch (opt) {
 		case 'i':
@@ -294,38 +336,60 @@ int main (int argc, char *argv[]) {
 			interface_name[sizeof(interface_name) - 1] = '\0';
 			break;
 		case 's':
-			rc = ringstatusget_do (interface_name);
+			action = ACTION_LINKSTATUS_GET;
+			break;
+		case 'b':
+			brief = 1;
 			break;
 		case 'R':
-			rc = reload_config_do ();
+			action = ACTION_RELOAD_CONFIG;
 			break;
-		case 'r':
-			ringreenable_do ();
+		case 'L':
+			action = ACTION_REOPEN_LOG_FILES;
 			break;
 		case 'k':
-			if (util_strtonum(optarg, 1, UINT_MAX, &l) == -1) {
-				fprintf(stderr, "The nodeid was not valid, try a positive number\n");
-				exit(EXIT_FAILURE);
-			}
-			nodeid = l;
-			killnode_do(nodeid);
+			nodeid = atoi (optarg);
+			action = ACTION_KILL_NODE;
 			break;
 		case 'H':
-			shutdown_do();
+			action = ACTION_SHUTDOW;
 			break;
 		case 'a':
-			if (util_strtonum(optarg, 1, UINT_MAX, &l) == -1) {
-				fprintf(stderr, "The nodeid was not valid, try a positive number\n");
-				exit(EXIT_FAILURE);
-			}
-			nodeid = l;
-			showaddrs_do(nodeid);
+			nodeid = atoi (optarg);
+			action = ACTION_SHOWADDR;
+			break;
+		case '?':
+			return (EXIT_FAILURE);
 			break;
 		case 'h':
-			usage_do();
+		default:
 			break;
 		}
 	}
+	switch(action) {
+	case ACTION_LINKSTATUS_GET:
+		rc = linkstatusget_do(interface_name, brief);
+		break;
+	case ACTION_RELOAD_CONFIG:
+		rc = reload_config_do();
+		break;
+	case ACTION_REOPEN_LOG_FILES:
+		rc = reopen_log_files_do();
+		break;
+	case ACTION_KILL_NODE:
+		killnode_do(nodeid);
+		break;
+	case ACTION_SHUTDOW:
+		shutdown_do();
+		break;
+	case ACTION_SHOWADDR:
+		showaddrs_do(nodeid);
+		break;
+	case ACTION_NOOP:
+	default:
+		usage_do();
+		break;
+	}
 
 	return (rc);
 }
